from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('sources', '0024_auto_20191219_0252'),
    ]

    operations = [
        migrations.DeleteModel(
            name='SourceLog',
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.dependencies.classes import BinaryDependency, PythonDependency

from .backends.literals import DEFAULT_FILE_PATH, DEFAULT_MIMETYPE_PATH


BinaryDependency(
    label='File::MimeInfo', help_text=_(
        'This module can be used to determine the mime type of a file. '
        'It tries to implement the freedesktop specification for a shared '
        'MIME database.'
    ), module=__name__, name='libfile-mimeinfo-perl',
    path=DEFAULT_MIMETYPE_PATH
)
BinaryDependency(
    label='file', help_text=_('determine file type using content tests'),
    module=__name__, name='file', path=DEFAULT_FILE_PATH
)
PythonDependency(
    copyright_text='''
        The MIT License (MIT)

        Copyright (c) 2001-2014 Adam Hupp

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.
    ''', module=__name__, name='python-magic', version_string='==0.4.24'
)



# -*- coding: utf-8 -*-

from django.core.exceptions import ValidationError

from mayan.apps.documents.models import DocumentType
from mayan.apps.documents.tests.literals import TEST_DOCUMENT_TYPE_2_LABEL
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..models import DocumentMetadata

from .literals import (
    TEST_DEFAULT_VALUE, TEST_LOOKUP_TEMPLATE, TEST_LOOKUP_VALUE_INCORRECT,
    TEST_LOOKUP_VALUE_CORRECT
)
from .mixins import MetadataTypeTestMixin


class MetadataTypeTestCase(
    DocumentTestMixin, MetadataTypeTestMixin, BaseTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()

        self._create_test_document_stub()
        self._create_test_metadata_type(add_test_document_type=True)

    def test_add_new_metadata_type_on_document_type_change(self):
        """
        When switching document types, add the required metadata of the new
        document type, the value to the default of the metadata type.
        """
        self.test_metadata_type.default = TEST_DEFAULT_VALUE
        self.test_metadata_type.save()

        self.test_document_type_2 = DocumentType.objects.create(
            label=TEST_DOCUMENT_TYPE_2_LABEL
        )

        self.test_document_type_2.metadata.create(
            metadata_type=self.test_metadata_type, required=True
        )

        self.test_document.document_type_change(document_type=self.test_document_type_2)

        self.assertEqual(self.test_document.metadata.count(), 1)
        self.assertEqual(
            self.test_document.metadata.first().value, TEST_DEFAULT_VALUE
        )

    def test_default(self):
        self.test_metadata_type.default = TEST_DEFAULT_VALUE
        self.test_metadata_type.save()

        document_metadata = DocumentMetadata(
            document=self.test_document, metadata_type=self.test_metadata_type
        )

        document_metadata.full_clean()
        document_metadata.save()

        self.assertEqual(
            getattr(
                self.test_document.metadata_value_of,
                self.test_metadata_type.name
            ), TEST_DEFAULT_VALUE
        )

    def test_default_missing(self):
        document_metadata = DocumentMetadata(
            document=self.test_document, metadata_type=self.test_metadata_type
        )

        document_metadata.full_clean()
        document_metadata.save()

        self.assertEqual(
            getattr(
                self.test_document.metadata_value_of,
                self.test_metadata_type.name
            ), None
        )

    def test_delete_metadata_type_present_assigned_as_document_metadata(self):
        # GitLab issue #753
        document_metadata = DocumentMetadata(
            document=self.test_document, metadata_type=self.test_metadata_type,
            value=TEST_DEFAULT_VALUE
        )

        document_metadata.full_clean()
        document_metadata.save()

        # Must not raise an error
        self.test_metadata_type.delete()

    def test_delete_metadata_value_on_document_type_change(self):
        """
        Delete the old document metadata whose types are not present in the
        new document type
        """
        document_metadata = DocumentMetadata(
            document=self.test_document, metadata_type=self.test_metadata_type,
            value=TEST_DEFAULT_VALUE
        )

        document_metadata.full_clean()
        document_metadata.save()

        self.test_document_type_2 = DocumentType.objects.create(
            label=TEST_DOCUMENT_TYPE_2_LABEL
        )

        self.test_document.document_type_change(document_type=self.test_document_type_2)

        self.assertEqual(self.test_document.metadata.count(), 0)

    def test_duplicate_metadata_value_on_document_type_change(self):
        """
        Delete the old document metadata whose types are not present in the
        new document type
        """
        document_metadata = DocumentMetadata(
            document=self.test_document, metadata_type=self.test_metadata_type,
            value=TEST_DEFAULT_VALUE
        )

        document_metadata.full_clean()
        document_metadata.save()

        self.test_document_type_2 = DocumentType.objects.create(
            label=TEST_DOCUMENT_TYPE_2_LABEL
        )

        self.test_document_type_2.metadata.create(
            metadata_type=self.test_metadata_type, required=True
        )

        self.test_document.document_type_change(document_type=self.test_document_type_2)

        self.assertEqual(self.test_document.metadata.count(), 1)
        self.assertEqual(
            self.test_document.metadata.first().value, TEST_DEFAULT_VALUE
        )
        self.assertEqual(
            self.test_document.metadata.first().metadata_type, self.test_metadata_type
        )

    def test_lookup_empty_optional(self):
        """
        Checks for GitLab issue #250
        Empty optional lookup metadata trigger validation error
        """
        self.test_metadata_type.lookup = TEST_LOOKUP_TEMPLATE
        self.test_metadata_type.save()

        document_metadata = DocumentMetadata(
            document=self.test_document, metadata_type=self.test_metadata_type
        )

        document_metadata.full_clean()
        document_metadata.save()

    def test_lookup_unicode(self):
        # Should NOT return a ValidationError, otherwise test fails
        self.test_metadata_type.lookup = '测试1,测试2,test1,test2'
        self.test_metadata_type.save()
        self.test_metadata_type.validate_value(document_type=None, value='测试1')

    def test_lookup_non_unicode(self):
        # Should NOT return a ValidationError, otherwise test fails
        self.test_metadata_type.lookup = 'test1,test2'
        self.test_metadata_type.save()
        self.test_metadata_type.validate_value(document_type=None, value='test1')

    def test_lookup_with_correct_value(self):
        self.test_metadata_type.lookup = TEST_LOOKUP_TEMPLATE
        self.test_metadata_type.save()

        document_metadata = DocumentMetadata(
            document=self.test_document, metadata_type=self.test_metadata_type,
            value=TEST_LOOKUP_VALUE_CORRECT
        )

        document_metadata.full_clean()
        document_metadata.save()

        self.assertEqual(
            getattr(
                self.test_document.metadata_value_of,
                self.test_metadata_type.name
            ), TEST_LOOKUP_VALUE_CORRECT
        )

    def test_lookup_with_incorrect_value(self):
        self.test_metadata_type.lookup = TEST_LOOKUP_TEMPLATE
        self.test_metadata_type.save()

        document_metadata = DocumentMetadata(
            document=self.test_document, metadata_type=self.test_metadata_type,
            value=TEST_LOOKUP_VALUE_INCORRECT
        )

        with self.assertRaises(expected_exception=ValidationError):
            # Should return error
            document_metadata.full_clean()
            document_metadata.save()

    def test_preserve_metadata_value_on_document_type_change(self):
        """
        Preserve the document metadata that is present in the
        old and new document types
        """
        document_metadata = DocumentMetadata(
            document=self.test_document, metadata_type=self.test_metadata_type,
            value=TEST_DEFAULT_VALUE
        )

        document_metadata.full_clean()
        document_metadata.save()

        self.test_document_type_2 = DocumentType.objects.create(
            label=TEST_DOCUMENT_TYPE_2_LABEL
        )

        self.test_document_type_2.metadata.create(metadata_type=self.test_metadata_type)

        self.test_document.document_type_change(document_type=self.test_document_type_2)

        self.assertEqual(self.test_document.metadata.count(), 1)
        self.assertEqual(
            self.test_document.metadata.first().value, TEST_DEFAULT_VALUE
        )
        self.assertEqual(
            self.test_document.metadata.first().metadata_type, self.test_metadata_type
        )

    def test_required_metadata(self):
        self.test_document_type.metadata.all().delete()

        self.assertFalse(
            self.test_metadata_type.get_required_for(self.test_document_type)
        )

        self.test_document_type.metadata.create(
            metadata_type=self.test_metadata_type, required=False
        )

        self.assertFalse(
            self.test_metadata_type.get_required_for(self.test_document_type)
        )

        self.test_document_type.metadata.all().delete()

        self.test_document_type.metadata.create(
            metadata_type=self.test_metadata_type, required=True
        )

        self.assertTrue(
            self.test_metadata_type.get_required_for(self.test_document_type)
        )

    def test_method_get_absolute_url(self):
        self._create_test_metadata_type()

        self.assertTrue(self.test_metadata_type.get_absolute_url())



from rest_framework.generics import get_object_or_404

from mayan.apps.acls.models import AccessControlList

from ..models.document_models import Document
from ..models.document_type_models import DocumentType


class ParentObjectDocumentAPIViewMixin:
    def get_document(self, permission=None):
        queryset = self.get_document_queryset()

        if not permission:
            permission = getattr(
                self, 'mayan_external_object_permissions', {}
            ).get(self.request.method, (None,))[0]

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs['document_id']
        )

    def get_document_queryset(self):
        return Document.valid.all()


class ParentObjectDocumentFileAPIViewMixin(ParentObjectDocumentAPIViewMixin):
    def get_document_file(self, permission=None):
        queryset = self.get_document_file_queryset()

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs['document_file_id']
        )

    def get_document_file_queryset(self):
        return self.get_document().files.all()


class ParentObjectDocumentFilePageAPIViewMixin(
    ParentObjectDocumentFileAPIViewMixin
):
    def get_document_file_page(self, permission=None):
        queryset = self.get_document_file_page_queryset()

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs['document_file_page_id']
        )

    def get_document_file_page_queryset(self):
        return self.get_document_file().pages.all()


class ParentObjectDocumentTypeAPIViewMixin:
    def get_document_type(self, permission=None):
        queryset = DocumentType.objects.all()

        if not permission:
            permission = getattr(
                self, 'mayan_external_object_permissions', {}
            ).get(self.request.method, (None,))[0]

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs['document_type_id']
        )


class ParentObjectDocumentVersionAPIViewMixin(ParentObjectDocumentAPIViewMixin):
    def get_document_version(self, permission=None):
        queryset = self.get_document_version_queryset()

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs['document_version_id']
        )

    def get_document_version_queryset(self):
        return self.get_document().versions.all()


class ParentObjectDocumentVersionPageAPIViewMixin(
    ParentObjectDocumentVersionAPIViewMixin
):
    def get_document_version_page(self, permission=None):
        queryset = self.get_document_version_page_queryset()

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs['document_version_page_id']
        )

    def get_document_version_page_queryset(self):
        return self.get_document_version().pages.all()



from mayan.apps.views.forms import DetailForm

from .models import Cache, CachePartition


class CacheDetailForm(DetailForm):
    class Meta:
        fields = ('defined_storage_name',)
        model = Cache


class CachePartitionDetailForm(DetailForm):
    class Meta:
        fields = ('cache', 'name')
        model = CachePartition



from datetime import timedelta

from django.utils.translation import ugettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_d

from .literals import (
    TASK_DOWNLOAD_FILE_STALE_INTERVAL, TASK_SHARED_UPLOADS_STALE_INTERVAL
)

queue_storage_periodic = CeleryQueue(
    label=_('Storage periodic'), name='storage_periodic', transient=True,
    worker=worker_d
)

queue_storage_periodic.add_task_type(
    dotted_path='mayan.apps.storage.tasks.task_shared_upload_stale_delete',
    label=_('Delete stale uploads'), name='task_shared_upload_stale_delete',
    schedule=timedelta(
        seconds=TASK_SHARED_UPLOADS_STALE_INTERVAL
    )
)
queue_storage_periodic.add_task_type(
    dotted_path='mayan.apps.storage.tasks.task_download_files_stale_delete',
    label=_('Delete stale download files'),
    name='task_download_files_stale_delete',
    schedule=timedelta(
        seconds=TASK_DOWNLOAD_FILE_STALE_INTERVAL
    )
)



from mayan.apps.appearance.classes import Icon

icon_group_roles = Icon(driver_name='fontawesome', symbol='user-secret')
icon_permission = Icon(driver_name='fontawesome', symbol='thumbs-up')
icon_role_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='user-secret',
    secondary_symbol='plus'
)
icon_role_delete_single = Icon(driver_name='fontawesome', symbol='times')
icon_role_delete_multiple = icon_role_delete_single
icon_role_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_role_groups = Icon(driver_name='fontawesome', symbol='users')
icon_role_list = Icon(driver_name='fontawesome', symbol='user-secret')
icon_role_permissions = Icon(driver_name='fontawesome', symbol='thumbs-up')



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('file_metadata', '0006_rename_documentversiondriverentry_model')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='documentfiledriverentry',
            options={
                'ordering': ('document_file', 'driver'),
                'verbose_name': 'Document file driver entry',
                'verbose_name_plural': 'Document file driver entries'
            }
        ),
        migrations.AlterField(
            model_name='documentfiledriverentry',
            name='document_file',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='file_metadata_drivers',
                to='documents.DocumentFile', verbose_name='Document file'
            )
        ),
        migrations.AlterField(
            model_name='filemetadataentry',
            name='document_file_driver_entry',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='entries',
                to='file_metadata.DocumentFileDriverEntry',
                verbose_name='Document file driver entry'
            )
        )
    ]



from django.conf.urls import url

from .views import (
    SettingEditView, SettingNamespaceDetailView, SettingNamespaceListView
)

urlpatterns = [
    url(
        regex=r'^namespaces/settings/(?P<setting_global_name>\w+)/edit/$',
        name='setting_edit_view', view=SettingEditView.as_view()
    ),
    url(
        regex=r'^namespaces/$', name='namespace_list',
        view=SettingNamespaceListView.as_view()
    ),
    url(
        regex=r'^namespaces/(?P<namespace_name>\w+)/$',
        name='namespace_detail', view=SettingNamespaceDetailView.as_view()
    )
]



def condition_is_current_user(context, resolved_object):
    if condition_user_is_authenticated(context=context, resolved_object=resolved_object):
        if 'user' in context:
            return resolved_object == context['user']


def condition_not_is_current_user(context, resolved_object):
    return condition_user_is_authenticated(
        context=context, resolved_object=resolved_object
    ) and not condition_is_current_user(
        context=context, resolved_object=resolved_object
    )


def condition_user_is_authenticated(context, resolved_object):
    if 'user' in context:
        return context['user'].is_authenticated



from mayan.apps.dependencies.classes import PythonDependency

PythonDependency(
    module=__name__, name='Whoosh', version_string='==2.7.4'
)
PythonDependency(
    module=__name__, name='elasticsearch', version_string='==7.16.0'
)
PythonDependency(
    module=__name__, name='elasticsearch-dsl', version_string='==7.4.0'
)



from django.db.models import Q

from ..models import SmartLink, SmartLinkCondition

from .literals import (
    TEST_SMART_LINK_CONDITION_FOREIGN_DOCUMENT_DATA,
    TEST_SMART_LINK_CONDITION_EXPRESSION,
    TEST_SMART_LINK_CONDITION_EXPRESSION_EDITED,
    TEST_SMART_LINK_CONDITION_INCLUSION, TEST_SMART_LINK_CONDITION_OPERATOR,
    TEST_SMART_LINK_DYNAMIC_LABEL, TEST_SMART_LINK_LABEL,
    TEST_SMART_LINK_LABEL_EDITED
)


class DocumentTypeAddRemoveSmartLinkViewTestMixin:
    def _request_test_document_type_smart_link_add_remove_get_view(self):
        return self.get(
            viewname='linking:document_type_smart_links', kwargs={
                'document_type_id': self.test_document_type.pk,
            }
        )

    def _request_test_document_type_smart_link_add_view(self):
        return self.post(
            viewname='linking:document_type_smart_links', kwargs={
                'document_type_id': self.test_document_type.pk,
            }, data={
                'available-submit': 'true',
                'available-selection': self.test_smart_link.pk
            }
        )

    def _request_test_document_type_smart_link_remove_view(self):
        return self.post(
            viewname='linking:document_type_smart_links', kwargs={
                'document_type_id': self.test_document_type.pk,
            }, data={
                'added-submit': 'true',
                'added-selection': self.test_smart_link.pk
            }
        )


class ResolvedSmartLinkAPIViewTestMixin:
    def _request_resolved_smart_link_detail_api_view(self):
        return self.get(
            viewname='rest_api:resolvedsmartlink-detail',
            kwargs={
                'document_id': self.test_documents[0].pk,
                'resolved_smart_link_id': self.test_smart_link.pk
            }
        )
        self._create_test_smart_link(add_test_document_type=True)

    def _request_resolved_smart_link_list_api_view(self):
        return self.get(
            viewname='rest_api:resolvedsmartlink-list', kwargs={
                'document_id': self.test_documents[0].pk
            }
        )

    def _request_resolved_smart_link_document_list_api_view(self):
        return self.get(
            viewname='rest_api:resolvedsmartlinkdocument-list',
            kwargs={
                'document_id': self.test_documents[0].pk,
                'resolved_smart_link_id': self.test_smart_link.pk
            }
        )


class SmartLinkAPIViewTestMixin:
    def _request_test_smart_link_create_api_view(self):
        pk_list = list(SmartLink.objects.values('pk'))

        response = self.post(
            viewname='rest_api:smartlink-list', data={
                'label': TEST_SMART_LINK_LABEL
            }
        )

        try:
            self.test_smart_link = SmartLink.objects.get(
                ~Q(pk__in=pk_list)
            )
        except SmartLink.DoesNotExist:
            self.test_smart_link = None

        return response

    def _request_test_smart_link_delete_api_view(self):
        return self.delete(
            viewname='rest_api:smartlink-detail', kwargs={
                'smart_link_id': self.test_smart_link.pk
            }
        )

    def _request_test_smart_link_detail_api_view(self):
        return self.get(
            viewname='rest_api:smartlink-detail', kwargs={
                'smart_link_id': self.test_smart_link.pk
            }
        )

    def _request_test_smart_link_edit_patch_api_view(self):
        return self.patch(
            viewname='rest_api:smartlink-detail',
            kwargs={'smart_link_id': self.test_smart_link.pk}, data={
                'label': TEST_SMART_LINK_LABEL_EDITED
            }
        )

    def _request_test_smart_link_edit_put_api_view(self):
        return self.put(
            viewname='rest_api:smartlink-detail',
            kwargs={'smart_link_id': self.test_smart_link.pk}, data={
                'label': TEST_SMART_LINK_LABEL_EDITED
            }
        )

    def _request_test_smart_link_list_api_view(self):
        return self.get(viewname='rest_api:smartlink-list')


class SmartLinkConditionAPIViewTestMixin:
    def _request_smart_link_condition_create_api_view(self):
        pk_list = list(SmartLinkCondition.objects.values('pk'))

        response = self.post(
            viewname='rest_api:smartlinkcondition-list',
            kwargs={'smart_link_id': self.test_smart_link.pk}, data={
                'foreign_document_data': TEST_SMART_LINK_CONDITION_FOREIGN_DOCUMENT_DATA,
                'expression': TEST_SMART_LINK_CONDITION_EXPRESSION,
                'operator': TEST_SMART_LINK_CONDITION_OPERATOR
            }
        )

        try:
            self.test_smart_link_condition = SmartLinkCondition.objects.get(
                ~Q(pk__in=pk_list)
            )
        except SmartLinkCondition.DoesNotExist:
            self.test_smart_link_condition = None

        return response

    def _request_smart_link_condition_delete_api_view(self):
        return self.delete(
            viewname='rest_api:smartlinkcondition-detail',
            kwargs={
                'smart_link_id': self.test_smart_link.pk,
                'smart_link_condition_id': self.test_smart_link_condition.pk
            }
        )

    def _request_smart_link_condition_detail_api_view(self):
        return self.get(
            viewname='rest_api:smartlinkcondition-detail',
            kwargs={
                'smart_link_id': self.test_smart_link.pk,
                'smart_link_condition_id': self.test_smart_link_condition.pk
            }
        )

    def _request_smart_link_condition_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:smartlinkcondition-detail',
            kwargs={
                'smart_link_id': self.test_smart_link.pk,
                'smart_link_condition_id': self.test_smart_link_condition.pk
            }, data={
                'expression': TEST_SMART_LINK_CONDITION_EXPRESSION_EDITED,
            }
        )

    def _request_smart_link_condition_edit_via_put_api_view(self):
        return self.put(
            viewname='rest_api:smartlinkcondition-detail',
            kwargs={
                'smart_link_id': self.test_smart_link.pk,
                'smart_link_condition_id': self.test_smart_link_condition.pk
            }, data={
                'expression': TEST_SMART_LINK_CONDITION_EXPRESSION_EDITED,
                'foreign_document_data': TEST_SMART_LINK_CONDITION_FOREIGN_DOCUMENT_DATA,
                'operator': TEST_SMART_LINK_CONDITION_OPERATOR

            }
        )

    def _request_smart_link_condition_list_api_view(self):
        return self.get(
            viewname='rest_api:smartlinkcondition-list', kwargs={
                'smart_link_id': self.test_smart_link.pk
            }
        )


class SmartLinkConditionViewTestMixin:
    def _request_test_smart_link_condition_create_view(self):
        pk_list = list(SmartLinkCondition.objects.values('pk'))

        response = self.post(
            viewname='linking:smart_link_condition_create', kwargs={
                'smart_link_id': self.test_smart_link.pk
            }, data={
                'inclusion': TEST_SMART_LINK_CONDITION_INCLUSION,
                'foreign_document_data': TEST_SMART_LINK_CONDITION_FOREIGN_DOCUMENT_DATA,
                'expression': TEST_SMART_LINK_CONDITION_EXPRESSION,
                'operator': TEST_SMART_LINK_CONDITION_OPERATOR
            }
        )

        try:
            self.test_smart_link_condition = SmartLinkCondition.objects.get(
                ~Q(pk__in=pk_list)
            )
        except SmartLinkCondition.DoesNotExist:
            self.test_smart_link_condition = None

        return response

    def _request_test_smart_link_condition_delete_view(self):
        return self.post(
            viewname='linking:smart_link_condition_delete', kwargs={
                'smart_link_condition_id': self.test_smart_link_condition.pk
            }
        )

    def _request_test_smart_link_condition_edit_view(self):
        return self.post(
            viewname='linking:smart_link_condition_edit', kwargs={
                'smart_link_condition_id': self.test_smart_link_condition.pk
            }, data={
                'inclusion': TEST_SMART_LINK_CONDITION_INCLUSION,
                'foreign_document_data': TEST_SMART_LINK_CONDITION_FOREIGN_DOCUMENT_DATA,
                'expression': TEST_SMART_LINK_CONDITION_EXPRESSION_EDITED,
                'operator': TEST_SMART_LINK_CONDITION_OPERATOR
            }
        )

    def _request_test_smart_link_condition_list_view(self):
        return self.get(
            viewname='linking:smart_link_condition_list', kwargs={
                'smart_link_id': self.test_smart_link.pk
            }
        )


class SmartLinkDocumentTypeAPIViewTestMixin:
    def _request_test_smart_link_document_type_add_api_view(self):
        return self.post(
            viewname='rest_api:smartlink-document_type-add',
            kwargs={'smart_link_id': self.test_smart_link.pk}, data={
                'document_type': self.test_document_type.pk
            }
        )

    def _request_test_smart_link_document_type_list_api_view(self):
        return self.get(
            viewname='rest_api:smartlink-document_type-list', kwargs={
                'smart_link_id': self.test_smart_link.pk,
            }
        )

    def _request_test_smart_link_document_type_remove_api_view(self):
        return self.post(
            viewname='rest_api:smartlink-document_type-remove',
            kwargs={'smart_link_id': self.test_smart_link.pk}, data={
                'document_type': self.test_document_type.pk
            }
        )


class SmartLinkDocumentTypeViewTestMixin:
    def _request_test_smart_link_document_type_add_remove_get_view(self):
        return self.get(
            viewname='linking:smart_link_document_types', kwargs={
                'smart_link_id': self.test_smart_link.pk,
            }
        )

    def _request_test_smart_link_document_type_add_view(self):
        return self.post(
            viewname='linking:smart_link_document_types', kwargs={
                'smart_link_id': self.test_smart_link.pk,
            }, data={
                'available-submit': 'true',
                'available-selection': self.test_document_type.pk
            }
        )

    def _request_test_smart_link_document_type_remove_view(self):
        return self.post(
            viewname='linking:smart_link_document_types', kwargs={
                'smart_link_id': self.test_smart_link.pk,
            }, data={
                'added-submit': 'true',
                'added-selection': self.test_document_type.pk
            }
        )


class ResolvedSmartLinkDocumentViewTestMixin:
    def _request_test_document_resolved_smart_link_document_list_view(self):
        return self.get(
            viewname='linking:smart_link_instance_view', kwargs={
                'document_id': self.test_documents[0].pk,
                'smart_link_id': self.test_smart_link.pk
            }
        )

    def _request_test_document_resolved_smart_link_list_view(self):
        return self.get(
            viewname='linking:smart_link_instances_for_document', kwargs={
                'document_id': self.test_documents[0].pk
            }
        )


class SmartLinkTestMixin:
    def _create_test_smart_link(self, add_test_document_type=False):
        self.test_smart_link = SmartLink.objects.create(
            label=TEST_SMART_LINK_LABEL,
            dynamic_label=TEST_SMART_LINK_DYNAMIC_LABEL
        )
        if add_test_document_type:
            self.test_smart_link.document_types.add(self.test_document_type)

    def _create_test_smart_link_condition(self):
        self.test_smart_link_condition = SmartLinkCondition.objects.create(
            smart_link=self.test_smart_link,
            foreign_document_data=TEST_SMART_LINK_CONDITION_FOREIGN_DOCUMENT_DATA,
            expression=TEST_SMART_LINK_CONDITION_EXPRESSION,
            operator=TEST_SMART_LINK_CONDITION_OPERATOR
        )


class SmartLinkViewTestMixin:
    def _request_test_smart_link_create_view(self):
        # Typecast to list to force queryset evaluation
        values = list(SmartLink.objects.values_list('pk', flat=True))

        response = self.post(
            viewname='linking:smart_link_create', data={
                'label': TEST_SMART_LINK_LABEL
            }
        )

        self.test_smart_link = SmartLink.objects.exclude(pk__in=values).first()

        return response

    def _request_test_smart_link_delete_view(self):
        return self.post(
            viewname='linking:smart_link_delete', kwargs={
                'smart_link_id': self.test_smart_link.pk
            }
        )

    def _request_test_smart_link_edit_view(self):
        return self.post(
            viewname='linking:smart_link_edit', kwargs={
                'smart_link_id': self.test_smart_link.pk
            }, data={
                'label': TEST_SMART_LINK_LABEL_EDITED
            }
        )

    def _request_test_smart_link_list_view(self):
        return self.get(viewname='linking:smart_link_list')



import os

from django.conf import settings
from django.db.models import Q

from mayan.apps.converter.classes import Layer

from ...literals import DOCUMENT_FILE_ACTION_PAGES_NEW, PAGE_RANGE_ALL
from ...models import Document, DocumentType

from ..literals import (
    DEFAULT_DOCUMENT_STUB_LABEL, TEST_DOCUMENT_DESCRIPTION,
    TEST_DOCUMENT_DESCRIPTION_EDITED, TEST_DOCUMENT_TYPE_LABEL,
    TEST_SMALL_DOCUMENT_FILENAME, TEST_SMALL_DOCUMENT_PATH
)


class DocumentAPIViewTestMixin:
    def _request_test_document_change_type_api_view(self):
        return self.post(
            viewname='rest_api:document-change-type', kwargs={
                'document_id': self.test_document.pk
            }, data={'document_type_id': self.test_document_types[1].pk}
        )

    def _request_test_document_create_api_view(self):
        pk_list = list(Document.objects.values_list('pk', flat=True))

        response = self.post(
            viewname='rest_api:document-list', data={
                'document_type_id': self.test_document_type.pk
            }
        )

        try:
            self.test_document = Document.objects.get(
                ~Q(pk__in=pk_list)
            )
        except Document.DoesNotExist:
            self.test_document = None

        return response

    def _request_test_document_detail_api_view(self):
        return self.get(
            viewname='rest_api:document-detail', kwargs={
                'document_id': self.test_document.pk
            }
        )

    def _request_test_document_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:document-detail', kwargs={
                'document_id': self.test_document.pk
            }, data={'description': TEST_DOCUMENT_DESCRIPTION_EDITED}
        )

    def _request_test_document_edit_via_put_api_view(self):
        return self.put(
            viewname='rest_api:document-detail', kwargs={
                'document_id': self.test_document.pk
            }, data={'description': TEST_DOCUMENT_DESCRIPTION_EDITED}
        )

    def _request_test_document_list_api_view(self):
        return self.get(viewname='rest_api:document-list')

    def _request_test_document_upload_api_view(self):
        pk_list = list(Document.objects.values_list('pk', flat=True))

        with open(file=TEST_SMALL_DOCUMENT_PATH, mode='rb') as file_object:
            response = self.post(
                viewname='rest_api:document-upload', data={
                    'document_type_id': self.test_document_type.pk,
                    'file': file_object
                }
            )

        try:
            self.test_document = Document.objects.get(
                ~Q(pk__in=pk_list)
            )
        except Document.DoesNotExist:
            self.test_document = None

        return response


class DocumentTestMixin:
    auto_create_test_document_type = True
    auto_upload_test_document = True
    test_document_file_filename = TEST_SMALL_DOCUMENT_FILENAME
    test_document_file_path = None
    test_document_filename = TEST_SMALL_DOCUMENT_FILENAME
    test_document_language = None
    test_document_path = None
    auto_delete_test_document_type = True

    def setUp(self):
        super().setUp()
        Layer.invalidate_cache()

        self.test_documents = []
        self.test_document_types = []

        if self.auto_create_test_document_type:
            self._create_test_document_type()

            if self.auto_upload_test_document:
                self._upload_test_document()

    def tearDown(self):
        if self.auto_delete_test_document_type:
            for document_type in DocumentType.objects.all():
                document_type.delete()
        super().tearDown()

    def _create_test_document_stub(self, document_type=None, label=None):
        self.test_document_stub = Document.objects.create(
            document_type=document_type or self.test_document_type,
            label=label or '{}_{}'.format(
                DEFAULT_DOCUMENT_STUB_LABEL, len(self.test_documents)
            )
        )
        self.test_document = self.test_document_stub
        self.test_documents.append(self.test_document)

    def _create_test_document_type(self, label=None):
        label = label or '{}_{}'.format(
            TEST_DOCUMENT_TYPE_LABEL, len(self.test_document_types)
        )

        self.test_document_type = DocumentType.objects.create(label=label)
        self.test_document_types.append(self.test_document_type)

    def _calculate_test_document_path(self):
        if not self.test_document_path:
            self.test_document_path = os.path.join(
                settings.BASE_DIR, 'apps', 'documents', 'tests', 'contrib',
                'sample_documents', self.test_document_filename
            )

    def _calculate_test_document_file_path(self):
        if not self.test_document_file_path:
            self.test_document_file_path = os.path.join(
                settings.BASE_DIR, 'apps', 'documents', 'tests', 'contrib',
                'sample_documents', self.test_document_file_filename
            )

    def _upload_test_document(
        self, description=None, document_file_attributes=None,
        document_type=None, document_version_attributes=None, label=None,
        _user=None
    ):
        self._calculate_test_document_path()

        if not label:
            label = self.test_document_filename

        test_document_description = description or '{}_{}'.format(
            TEST_DOCUMENT_DESCRIPTION, len(self.test_documents)
        )

        document_type = document_type or self.test_document_type

        with open(file=self.test_document_path, mode='rb') as file_object:
            document, document_file = document_type.new_document(
                description=test_document_description,
                file_object=file_object, label=label,
                language=self.test_document_language, _user=_user
            )

        self.test_document = document
        self.test_documents.append(document)

        self.test_document_file = document_file
        self.test_document_file_page = document_file.file_pages.first()
        self.test_document_version = self.test_document.version_active
        self.test_document_version_page = self.test_document_version.version_pages.first()

        if document_file_attributes:
            for key, value in document_file_attributes.items():
                setattr(self.test_document_file, key, value)

            self.test_document_file.save()

        if document_version_attributes:
            for key, value in document_version_attributes.items():
                setattr(self.test_document_version, key, value)

            self.test_document_version.save()

    def _upload_test_document_file(self, action=None, _user=None):
        self._calculate_test_document_file_path()

        if not action:
            action = DOCUMENT_FILE_ACTION_PAGES_NEW

        with open(file=self.test_document_path, mode='rb') as file_object:
            self.test_document_file = self.test_document.file_new(
                action=action, file_object=file_object, _user=_user
            )

        self.test_document_file_page = self.test_document_file.pages.first()
        self.test_document_version = self.test_document.version_active


class DocumentViewTestMixin:
    def _request_test_document_list_view(self):
        return self.get(viewname='documents:document_list')

    def _request_test_document_preview_view(self):
        return self.get(
            viewname='documents:document_preview', kwargs={
                'document_id': self.test_document.pk
            }
        )

    def _request_test_document_print_form_view(self):
        return self.get(
            viewname='documents:document_print_form', kwargs={
                'document_id': self.test_document.pk,
            }, data={
                'page_group': PAGE_RANGE_ALL
            }
        )

    def _request_test_document_print_view(self):
        return self.get(
            viewname='documents:document_print_view', kwargs={
                'document_id': self.test_document.pk,
            }, query={
                'page_group': PAGE_RANGE_ALL
            }
        )

    def _request_test_document_properties_edit_get_view(self):
        return self.get(
            viewname='documents:document_properties_edit', kwargs={
                'document_id': self.test_document.pk
            }
        )

    def _request_test_document_properties_view(self):
        return self.get(
            viewname='documents:document_properties', kwargs={
                'document_id': self.test_document.pk
            }
        )

    def _request_test_document_type_change_get_view(self):
        return self.get(
            viewname='documents:document_type_change', kwargs={
                'document_id': self.test_document.pk
            }
        )

    def _request_test_document_type_change_post_view(self):
        return self.post(
            viewname='documents:document_type_change', kwargs={
                'document_id': self.test_document.pk
            }, data={'document_type': self.test_document_types[1].pk}
        )

    def _request_test_document_multiple_type_change(self):
        return self.post(
            viewname='documents:document_multiple_type_change',
            data={
                'id_list': self.test_document.pk,
                'document_type': self.test_document_types[1].pk
            }
        )



from mayan.apps.rest_api.tests.base import BaseAPITestCase
from mayan.apps.testing.tests.base import (
    BaseTestCase, BaseTransactionTestCase, GenericViewTestCase,
    GenericTransactionViewTestCase
)


from .mixins.document_mixins import DocumentTestMixin


class GenericDocumentTestCase(DocumentTestMixin, BaseTestCase):
    """Base test case when testing document models or classes."""


class GenericDocumentAPIViewTestCase(DocumentTestMixin, BaseAPITestCase):
    """Base test case when testing document API views."""


class GenericDocumentViewTestCase(DocumentTestMixin, GenericViewTestCase):
    """Base test case when testing document views."""


class GenericTransactionDocumentTestCase(
    DocumentTestMixin, BaseTransactionTestCase
):
    """
    Base test case when testing document models or classes with transactions.
    """


class GenericTransactionDocumentViewTestCase(
    DocumentTestMixin, GenericTransactionViewTestCase
):
    """Base test case when testing document views with transactions."""



import logging

from django.apps import apps

from mayan.apps.lock_manager.backends.base import LockingBackend
from mayan.apps.lock_manager.exceptions import LockError
from mayan.celery import app

from .literals import CHECKOUT_EXPIRATION_LOCK_EXPIRE

logger = logging.getLogger(name=__name__)


@app.task(ignore_result=True)
def task_check_expired_check_outs():
    DocumentCheckout = apps.get_model(
        app_label='checkouts', model_name='DocumentCheckout'
    )

    logger.debug(msg='executing...')
    lock_id = 'task_expired_check_outs'
    try:
        logger.debug('trying to acquire lock: %s', lock_id)
        lock = LockingBackend.get_backend().acquire_lock(
            name=lock_id, timeout=CHECKOUT_EXPIRATION_LOCK_EXPIRE
        )
        logger.debug('acquired lock: %s', lock_id)
        DocumentCheckout.objects.check_in_expired_check_outs()
        lock.release()
    except LockError:
        logger.debug(msg='unable to obtain lock')



from unittest import skip

from django.db import models

from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.search import document_search
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..classes import SearchModel

from .mixins import SearchTestMixin


class CommonBackendFunctionalityTestCaseMixin(SearchTestMixin):
    def _create_test_instances(self):
        self._test_object_grandparent = self.TestModelGrandParent.objects.create(
            label='grandparent'
        )
        self._test_object_parent = self.TestModelParent.objects.create(
            parent=self._test_object_grandparent,
            label='parent'
        )
        self._test_object_grandchild = self.TestModelGrandChild.objects.create(
            parent=self._test_object_parent,
            label='grandchild'
        )
        self._test_object_grandchild_proxy = self.TestModelGrandChildProxy.objects.first()

        self._test_object_attribute = self.TestModelAttribute.objects.create(
            label='attribute'
        )
        self._test_object_grandchild.attributes.add(self._test_object_attribute)

    def _create_test_models(self):
        self.TestModelAttribute = self._create_test_model(
            fields={
                'label': models.CharField(
                    max_length=32
                )
            }, model_name='TestModelAttribute'
        )
        self.TestModelGrandParent = self._create_test_model(
            fields={
                'label': models.CharField(
                    max_length=32
                )
            }, model_name='TestModelGrandParent'
        )
        self.TestModelParent = self._create_test_model(
            fields={
                'parent': models.ForeignKey(
                    on_delete=models.CASCADE, related_name='children',
                    to='TestModelGrandParent',
                ),
                'label': models.CharField(
                    max_length=32
                )
            }, model_name='TestModelParent'
        )
        self.TestModelGrandChild = self._create_test_model(
            fields={
                'parent': models.ForeignKey(
                    on_delete=models.CASCADE, related_name='children',
                    to='TestModelParent',
                ),
                'attributes': models.ManyToManyField(
                    related_name='children', to='TestModelAttribute',
                ),
                'label': models.CharField(
                    max_length=32
                )
            }, model_name='TestModelGrandChild'
        )
        self.TestModelGrandChildProxy = self._create_test_model(
            base_class=self.TestModelGrandChild,
            model_name='TestModelGrandChildProxy',
            options={
                'proxy': True
            }
        )

    def _setup_test_model_search(self):
        self._test_search_grandparent = SearchModel(
            app_label=self.TestModelGrandParent._meta.app_label,
            model_name=self.TestModelGrandParent._meta.model_name
        )

        self._test_search_grandparent.add_model_field(field='label')
        self._test_search_grandparent.add_model_field(
            field='children__label'
        )
        self._test_search_grandparent.add_model_field(
            field='children__children__label'
        )
        self._test_search_grandparent.add_model_field(
            field='children__children__attributes__label'
        )

        self._test_search_grandchild = SearchModel(
            app_label=self.TestModelGrandChild._meta.app_label,
            model_name=self.TestModelGrandChild._meta.model_name
        )
        self._test_search_grandchild.add_model_field(
            field='label'
        )
        self._test_search_grandchild.add_model_field(
            field='attributes__label'
        )
        self._test_search_grandchild.add_proxy_model(
            app_label=self.TestModelAttribute._meta.app_label,
            model_name='TestModelGrandChildProxy'
        )

        self._test_search_attribute = SearchModel(
            app_label=self.TestModelAttribute._meta.app_label,
            model_name=self.TestModelAttribute._meta.model_name
        )
        self._test_search_grandchild.add_model_field(
            field='label'
        )
        self._test_search_attribute.add_model_field(
            field='children__label'
        )

    def setUp(self):
        super().setUp()
        self._create_test_instances()

    # Direct object.

    def test_object_field_search(self):
        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'label': self._test_object_grandparent.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent in queryset)

    def test_object_field_update_search(self):
        old_label = self._test_object_grandparent.label
        self._test_object_grandparent.label = 'edited'
        self._test_object_grandparent.save()

        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'label': self._test_object_grandparent.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent in queryset)

        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'label': '"{}"'.format(old_label)
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent not in queryset)

    def test_object_field_delete_search(self):
        self._test_object_grandparent.delete()

        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'label': self._test_object_grandparent.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent not in queryset)

    # Direct object many to many.

    def test_object_many_to_many_search(self):
        queryset = self.search_backend.search(
            search_model=self._test_search_grandchild,
            query={
                'attributes__label': self._test_object_attribute.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandchild in queryset)

    def test_object_many_to_many_delete_search(self):
        self._test_object_attribute.delete()

        queryset = self.search_backend.search(
            search_model=self._test_search_grandchild,
            query={
                'attributes__label': self._test_object_attribute.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandchild not in queryset)

    def test_object_many_to_many_updated_search(self):
        old_label_value = self._test_object_attribute.label
        self._test_object_attribute.label = 'edited'
        self._test_object_attribute.save()

        queryset = self.search_backend.search(
            search_model=self._test_search_grandchild,
            query={
                'attributes__label': self._test_object_attribute.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandchild in queryset)

        queryset = self.search_backend.search(
            search_model=self._test_search_grandchild,
            query={
                'attributes__label': old_label_value
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandchild not in queryset)

    def test_object_many_to_many_remove_search(self):
        self._test_object_grandchild.attributes.remove(self._test_object_attribute)

        queryset = self.search_backend.search(
            search_model=self._test_search_grandchild,
            query={
                'attributes__label': self._test_object_attribute.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandchild not in queryset)

    def test_object_reverse_many_to_many_search(self):
        queryset = self.search_backend.search(
            search_model=self._test_search_attribute,
            query={
                'children__label': self._test_object_grandchild.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_attribute in queryset)

    def test_object_reverse_many_to_many_parent_delete_search(self):
        self._test_object_grandchild.delete()

        queryset = self.search_backend.search(
            search_model=self._test_search_attribute,
            query={
                'children__label': self._test_object_grandchild.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_attribute not in queryset)

    # Related object

    def test_object_related_search(self):
        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'children__label': self._test_object_parent.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent in queryset)

    def test_object_related_delete_search(self):
        self._test_object_parent.delete()

        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'children__label': self._test_object_parent.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent not in queryset)

    def test_object_related_update_search(self):
        old_label_value = self._test_object_parent.label
        self._test_object_parent.label = 'edited'
        self._test_object_parent.save()

        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'children__label': self._test_object_parent.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent in queryset)

        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'children__label': old_label_value
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent not in queryset)

    def test_object_related_multiple_level_search(self):
        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'children__children__label': self._test_object_grandchild.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent in queryset)

    def test_object_related_multiple_level_delete_search(self):
        self._test_object_grandchild.delete()

        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'children__children__label': self._test_object_grandchild.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent not in queryset)

    def test_object_related_multiple_level_update_search(self):
        old_label_value = self._test_object_grandchild.label
        self._test_object_grandchild.label = 'edited'
        self._test_object_grandchild.save()

        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'children__children__label': self._test_object_grandchild.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent in queryset)

        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'children__children__label': old_label_value
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent not in queryset)

    # Related with many to many

    def test_object_related_multiple_level_many_to_many_search(self):
        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'children__children__attributes__label': self._test_object_attribute.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent in queryset)

    def test_object_related_multiple_level_many_to_many_delete_search(self):
        self._test_object_attribute.delete()

        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'children__children__attributes__label': self._test_object_attribute.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent not in queryset)

    def test_object_related_multiple_level_many_to_many_updated_search(self):
        old_label_value = self._test_object_attribute.label
        self._test_object_attribute.label = 'edited'
        self._test_object_attribute.save()

        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'children__children__attributes__label': self._test_object_attribute.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent in queryset)

        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'children__children__attributes__label': old_label_value
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent not in queryset)

    def test_object_related_multiple_level_many_to_many_remove_search(self):
        self._test_object_grandchild.attributes.remove(self._test_object_attribute)

        queryset = self.search_backend.search(
            search_model=self._test_search_grandparent,
            query={
                'children__children__attributes__label': self._test_object_attribute.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandparent not in queryset)

    # Proxies

    def test_proxy_object_field_update_search(self):
        old_label = self._test_object_grandchild_proxy.label
        self._test_object_grandchild_proxy.label = 'edited'
        self._test_object_grandchild_proxy.save()

        queryset = self.search_backend.search(
            search_model=self._test_search_grandchild,
            query={
                'label': self._test_object_grandchild_proxy.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandchild in queryset)

        queryset = self.search_backend.search(
            search_model=self._test_search_grandchild,
            query={
                'label': '"{}"'.format(old_label)
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandchild not in queryset)

    def test_proxy_object_delete_search(self):
        self._test_object_grandchild_proxy.delete()

        queryset = self.search_backend.search(
            search_model=self._test_search_grandchild,
            query={
                'label': self._test_object_grandchild.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandchild not in queryset)

    def test_proxy_object_many_to_many_remove_search(self):
        self._test_object_grandchild_proxy.attributes.remove(
            self._test_object_attribute
        )

        queryset = self.search_backend.search(
            search_model=self._test_search_grandchild,
            query={
                'attributes__label': self._test_object_attribute.label
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandchild not in queryset)

    def test_undefined_search_field(self):
        queryset = self.search_backend.search(
            search_model=self._test_search_grandchild,
            query={
                'invalid': 'invalid'
            }, user=self._test_case_user
        )

        self.assertTrue(self._test_object_grandchild not in queryset)


class DjangoSearchBackendDocumentSearchTestCase(
    CommonBackendFunctionalityTestCaseMixin, DocumentTestMixin,
    BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'
    auto_upload_test_document = False

    def test_meta_only(self):
        self._upload_test_document(label='first_doc')
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'q': 'OR first'}, user=self._test_case_user
        )

        self.assertEqual(queryset.count(), 1)

    def test_simple_or_search(self):
        self._upload_test_document(label='first_doc')
        self.grant_access(
            obj=self.test_documents[0], permission=permission_document_view
        )
        self._upload_test_document(label='second_doc')
        self.grant_access(
            obj=self.test_documents[1], permission=permission_document_view
        )
        queryset = self.search_backend.search(
            search_model=document_search,
            query={'q': 'first OR second'}, user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 2)
        self.assertTrue(self.test_documents[0] in queryset)
        self.assertTrue(self.test_documents[1] in queryset)

    def test_advanced_or_search(self):
        self._upload_test_document(label='first_doc')
        self.grant_access(
            obj=self.test_documents[0], permission=permission_document_view
        )

        self._upload_test_document(label='second_doc')
        self.grant_access(
            obj=self.test_documents[1], permission=permission_document_view
        )

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'label': 'first OR second'},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 2)
        self.assertTrue(self.test_documents[0] in queryset)
        self.assertTrue(self.test_documents[1] in queryset)

    def test_simple_and_search(self):
        self._upload_test_document(label='second_doc')

        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'q': 'non_valid second'},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'q': 'second non_valid'},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)

    def test_simple_negated_search(self):
        self._upload_test_document(label='second_doc')

        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'q': '-non_valid second'},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 1)

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'label': '-second'},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'label': '-second -Mayan'},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'label': '-second OR -Mayan'},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 1)

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'label': '-non_valid -second'},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)

    def test_search_with_dashed_content(self):
        self._upload_test_document(label='second-document')

        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'label': '-second-document'},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'label': '-"second-document"'},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)


@skip('Skip until a Mock ElasticSearch server class is added.')
class ElasticSearchBackendDocumentSearchTestCase(
    CommonBackendFunctionalityTestCaseMixin, DocumentTestMixin,
    BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.elasticsearch.ElasticSearchBackend'
    auto_upload_test_document = False

    def test_simple_document_search(self):
        self._upload_test_document(label='first_doc')

        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'q': 'first*'}, user=self._test_case_user
        )

        self.assertEqual(queryset.count(), 1)
        self.assertTrue(self.test_document in queryset)


class WhooshSearchBackendDocumentSearchTestCase(
    CommonBackendFunctionalityTestCaseMixin, DocumentTestMixin,
    BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend'
    auto_upload_test_document = False

    def test_simple_search(self):
        self._upload_test_document(label='first_doc')

        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'q': 'first*'}, user=self._test_case_user
        )

        self.assertEqual(queryset.count(), 1)
        self.assertTrue(self.test_document in queryset)

    def test_meta_only(self):
        self._upload_test_document(label='first_doc')
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'q': 'OR first*'}, user=self._test_case_user
        )

        self.assertEqual(queryset.count(), 1)

    def test_simple_or_search(self):
        self._upload_test_document(label='first_doc')
        self.grant_access(
            obj=self.test_documents[0], permission=permission_document_view
        )
        self._upload_test_document(label='second_doc')
        self.grant_access(
            obj=self.test_documents[1], permission=permission_document_view
        )
        queryset = self.search_backend.search(
            search_model=document_search,
            query={'q': 'first* OR second*'}, user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 2)
        self.assertTrue(self.test_documents[0] in queryset)
        self.assertTrue(self.test_documents[1] in queryset)

    def test_advanced_or_search(self):
        self._upload_test_document(label='first_doc')
        self.grant_access(
            obj=self.test_documents[0], permission=permission_document_view
        )

        self._upload_test_document(label='second_doc')
        self.grant_access(
            obj=self.test_documents[1], permission=permission_document_view
        )

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'label': 'first* OR second*'},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 2)
        self.assertTrue(self.test_documents[0] in queryset)
        self.assertTrue(self.test_documents[1] in queryset)

    def test_simple_and_search(self):
        self._upload_test_document(label='second_doc')

        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'q': 'non_valid AND second*'},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)

        queryset = self.search_backend.search(
            search_model=document_search,
            query={'q': 'second* AND non_valid'},
            user=self._test_case_user
        )
        self.assertEqual(queryset.count(), 0)



import copy

from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.links import link_acl_list
from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_document_cabinet_add, icon_document_cabinet_remove,
    icon_document_multiple_cabinet_add, icon_cabinet_child_add,
    icon_cabinet_create, icon_cabinet_delete, icon_cabinet_edit,
    icon_cabinet_list, icon_cabinet_view
)
from .permissions import (
    permission_cabinet_add_document, permission_cabinet_create,
    permission_cabinet_delete, permission_cabinet_edit,
    permission_cabinet_view, permission_cabinet_remove_document
)


def condition_cabinet_is_root(context, resolved_object):
    return context['resolved_object'].is_root_node()


# Document links

link_document_cabinet_list = Link(
    args='resolved_object.pk', icon=icon_cabinet_list,
    permissions=(permission_cabinet_view,), text=_('Cabinets'),
    view='cabinets:document_cabinet_list'
)
link_document_cabinet_remove = Link(
    args='resolved_object.pk', icon=icon_document_cabinet_remove,
    permissions=(permission_cabinet_remove_document,),
    text=_('Remove from cabinets'), view='cabinets:document_cabinet_remove'
)
link_document_cabinet_add = Link(
    args='object.pk', icon=icon_document_cabinet_add,
    permissions=(permission_cabinet_add_document,), text=_('Add to cabinets'),
    view='cabinets:document_cabinet_add'
)
link_document_multiple_cabinet_add = Link(
    icon=icon_document_multiple_cabinet_add, text=_('Add to cabinets'),
    view='cabinets:document_multiple_cabinet_add'
)
link_multiple_document_cabinet_remove = Link(
    icon=icon_document_cabinet_remove, text=_('Remove from cabinets'),
    view='cabinets:multiple_document_cabinet_remove'
)

# Cabinet links

link_custom_acl_list = copy.copy(link_acl_list)
link_custom_acl_list.condition = condition_cabinet_is_root

link_cabinet_child_add = Link(
    args='object.pk', icon=icon_cabinet_child_add,
    permissions=(permission_cabinet_create,), text=_('Add new level'),
    view='cabinets:cabinet_child_add'
)
link_cabinet_create = Link(
    icon=icon_cabinet_create, permissions=(permission_cabinet_create,),
    text=_('Create cabinet'), view='cabinets:cabinet_create'
)
link_cabinet_delete = Link(
    args='object.pk', icon=icon_cabinet_delete,
    permissions=(permission_cabinet_delete,), tags='dangerous',
    text=_('Delete'), view='cabinets:cabinet_delete'
)
link_cabinet_edit = Link(
    args='object.pk', icon=icon_cabinet_edit,
    permissions=(permission_cabinet_edit,), text=_('Edit'),
    view='cabinets:cabinet_edit'
)
link_cabinet_list = Link(
    condition=factory_condition_queryset_access(
        app_label='cabinets', model_name='Cabinet',
        object_permission=permission_cabinet_view,
    ), icon=icon_cabinet_list,
    text=_('All'), view='cabinets:cabinet_list'
)
link_cabinet_view = Link(
    args='object.pk', icon=icon_cabinet_view,
    permissions=(permission_cabinet_view,), text=_('Details'),
    view='cabinets:cabinet_view'
)



from ...source_backends.literals import SOURCE_UNCOMPRESS_CHOICE_NEVER
from ...source_backends.sane_scanner_backends import SourceBackendSANEScanner

from .base_mixins import SourceTestMixin


class SANEScannerSourceTestMixin(SourceTestMixin):
    _create_source_method = '_create_test_sane_scanner_source'

    def _create_test_sane_scanner_source(self, extra_data=None):
        backend_data = {
            'device_name': 'test',
            'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER,
            'arguments': '{test-picture: grid}'
        }

        if extra_data:
            backend_data.update(extra_data)

        self._create_test_source(
            backend_path=SourceBackendSANEScanner.get_class_path(),
            backend_data=backend_data
        )



import json

from django.db.models import Q

from mayan.apps.documents.literals import DOCUMENT_FILE_ACTION_PAGES_NEW
from mayan.apps.documents.tests.literals import TEST_SMALL_DOCUMENT_PATH

from ...forms import NewDocumentForm
from ...models import Source
from ...source_backends.literals import (
    DEFAULT_PERIOD_INTERVAL, SOURCE_UNCOMPRESS_CHOICE_NEVER
)

from ..literals import (
    TEST_SOURCE_BACKEND_PATH, TEST_SOURCE_BACKEND_PERIODIC_PATH,
    TEST_SOURCE_LABEL, TEST_SOURCE_LABEL_EDITED
)
from ..mocks import MockRequest


class DocumentFileUploadViewTestMixin:
    def _request_document_file_upload_view(self):
        with open(file=TEST_SMALL_DOCUMENT_PATH, mode='rb') as file_object:
            return self.post(
                viewname='sources:document_file_upload', kwargs={
                    'document_id': self.test_document.pk,
                    'source_id': self.test_source.pk,
                }, data={
                    'document-action': DOCUMENT_FILE_ACTION_PAGES_NEW,
                    'source-file': file_object
                }
            )

    def _request_document_file_upload_no_source_view(self):
        with open(file=TEST_SMALL_DOCUMENT_PATH, mode='rb') as file_object:
            return self.post(
                viewname='sources:document_file_upload', kwargs={
                    'document_id': self.test_document.pk,
                }, data={
                    'document-action': DOCUMENT_FILE_ACTION_PAGES_NEW,
                    'source-file': file_object
                }
            )


class DocumentUploadWizardViewTestMixin:
    def _request_upload_interactive_view(self):
        return self.get(
            viewname='sources:document_upload_interactive', data={
                'document_type_id': self.test_document_type.pk,
            }
        )

    def _request_upload_wizard_view(self, document_path=TEST_SMALL_DOCUMENT_PATH):
        with open(file=document_path, mode='rb') as file_object:
            return self.post(
                viewname='sources:document_upload_interactive', kwargs={
                    'source_id': self.test_source.pk
                }, data={
                    'source-file': file_object,
                    'document_type_id': self.test_document_type.pk,
                }
            )


class InteractiveSourceBackendTestMixin:
    class MockSourceForm:
        def __init__(self, **kwargs):
            self.cleaned_data = kwargs

    def setUp(self):
        super().setUp()
        self.test_document_form = self.get_test_document_form()

    def get_test_document_form(self):
        document_form = NewDocumentForm(
            data={}, document_type=self.test_document_type
        )
        document_form.full_clean()

        return document_form

    def get_test_request(self):
        return MockRequest(user=self._test_case_user)


class SourceAPIViewTestMixin:
    def _request_test_source_create_api_view(
        self, backend_path=None, extra_data=None
    ):
        pk_list = list(Source.objects.values_list('pk', flat=True))

        data = {
            'backend_path': backend_path or TEST_SOURCE_BACKEND_PATH,
            'enabled': True,
            'label': TEST_SOURCE_LABEL
        }

        if extra_data:
            data.update(extra_data)

        response = self.post(viewname='rest_api:source-list', data=data)

        try:
            self.test_source = Source.objects.get(~Q(pk__in=pk_list))
        except Source.DoesNotExist:
            self.test_source = None

        return response

    def _request_test_source_delete_api_view(self):
        return self.delete(
            viewname='rest_api:source-detail', kwargs={
                'source_id': self.test_source.pk
            }
        )

    def _request_test_source_edit_api_view_via_patch(self):
        return self.patch(
            viewname='rest_api:source-detail', kwargs={
                'source_id': self.test_source.pk
            }, data={'label': TEST_SOURCE_LABEL_EDITED}
        )

    def _request_test_source_edit_api_view_via_put(self):
        data = {
            'backend_path': self.test_source.backend_path,
            'enabled': self.test_source.enabled,
            'label': TEST_SOURCE_LABEL_EDITED
        }

        return self.put(
            viewname='rest_api:source-detail', kwargs={
                'source_id': self.test_source.pk
            }, data=data
        )

    def _request_test_source_list_api_view(self):
        return self.get(viewname='rest_api:source-list')


class SourceTestMixin:
    _create_source_method = '_create_test_source'
    auto_create_test_source = True

    def setUp(self):
        super().setUp()
        self._test_sources = []

        if self.auto_create_test_source:
            getattr(self, self._create_source_method)()

    def _create_test_source(self, backend_path=None, backend_data=None):
        total_test_sources = len(self._test_sources)
        label = '{}_{}'.format(TEST_SOURCE_LABEL, total_test_sources)

        self.test_source = Source.objects.create(
            backend_path=backend_path or TEST_SOURCE_BACKEND_PATH,
            backend_data=json.dumps(obj=backend_data or {}),
            label=label
        )
        self._test_sources.append(self.test_source)


class PeriodicSourceBackendTestMixin(SourceTestMixin):
    _create_source_method = '_create_test_periodic_source_backend'

    def _create_test_periodic_source_backend(self, extra_data=None):
        backend_data = {
            'interval': DEFAULT_PERIOD_INTERVAL
        }

        if extra_data:
            backend_data.update(extra_data)

        self._create_test_source(
            backend_path=TEST_SOURCE_BACKEND_PERIODIC_PATH,
            backend_data=backend_data
        )


class SourceViewTestMixin:
    def _request_test_source_backend_selection_view(self):
        return self.get(
            viewname='sources:source_backend_selection'
        )

    def _request_test_source_create_view(
        self, backend_path=None, extra_data=None
    ):
        pk_list = list(Source.objects.values_list('pk', flat=True))

        data = {
            'enabled': True,
            'label': TEST_SOURCE_LABEL,
            'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER
        }

        if extra_data:
            data.update(extra_data)

        response = self.post(
            kwargs={
                'backend_path': backend_path or TEST_SOURCE_BACKEND_PATH
            }, viewname='sources:source_create', data=data
        )

        try:
            self.test_source = Source.objects.get(~Q(pk__in=pk_list))
        except Source.DoesNotExist:
            self.test_source = None

        return response

    def _request_test_source_delete_view(self):
        return self.post(
            viewname='sources:source_delete', kwargs={
                'source_id': self.test_source.pk
            }
        )

    def _request_test_source_edit_view(self):
        return self.post(
            viewname='sources:source_edit', kwargs={
                'source_id': self.test_source.pk
            }, data={
                'label': TEST_SOURCE_LABEL_EDITED
            }
        )

    def _request_test_source_list_view(self):
        return self.get(viewname='sources:source_list')

    def _request_test_source_test_get_view(self):
        return self.get(
            viewname='sources:source_test', kwargs={
                'source_id': self.test_source.pk
            }
        )

    def _request_test_source_test_post_view(self):
        return self.post(
            viewname='sources:source_test', kwargs={
                'source_id': self.test_source.pk
            }
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(label=_('Announcements'), name='announcement')

event_announcement_created = namespace.add_event_type(
    label=_('Announcement created'), name='announcement_created'
)
event_announcement_edited = namespace.add_event_type(
    label=_('Announcement edited'), name='announcement_edited'
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import (
    icon_entry_list, icon_document_type_ocr_settings,
    icon_document_type_submit,
    icon_document_version_ocr_content_delete_multiple,
    icon_document_version_ocr_content_delete_single,
    icon_document_version_ocr_content_view,
    icon_document_version_ocr_download,
    icon_document_version_ocr_errors_list,
    icon_document_version_ocr_submit_multiple,
    icon_document_version_ocr_submit_single,
    icon_document_version_page_ocr_content_detail_view,
    icon_document_version_page_ocr_content_edit_view
)
from .permissions import (
    permission_document_type_ocr_setup, permission_document_version_ocr,
    permission_document_version_ocr_content_edit,
    permission_document_version_ocr_content_view
)

# Document type

link_document_type_ocr_settings = Link(
    args='resolved_object.id',
    icon=icon_document_type_ocr_settings,
    permissions=(permission_document_type_ocr_setup,), text=_('Setup OCR'),
    view='ocr:document_type_ocr_settings'
)
link_document_type_submit = Link(
    icon=icon_document_type_submit,
    permissions=(permission_document_version_ocr,),
    text=_('OCR documents per type'), view='ocr:document_type_submit'
)

# Document version

link_document_version_ocr_content_delete_multiple = Link(
    icon=icon_document_version_ocr_content_delete_multiple,
    text=_('Delete OCR content'),
    view='ocr:document_version_ocr_content_delete_multiple'
)
link_document_version_ocr_content_delete_single = Link(
    args='resolved_object.id',
    icon=icon_document_version_ocr_content_delete_single,
    permissions=(permission_document_version_ocr,),
    text=_('Delete OCR content'),
    view='ocr:document_version_ocr_content_delete_single'
)
link_document_version_ocr_content_view = Link(
    args='resolved_object.id',
    icon=icon_document_version_ocr_content_view,
    permissions=(permission_document_version_ocr_content_view,),
    text=_('OCR'), view='ocr:document_version_ocr_content_view'
)
link_document_version_ocr_download = Link(
    args='resolved_object.id', icon=icon_document_version_ocr_download,
    permissions=(permission_document_version_ocr_content_view,),
    text=_('Download OCR text'), view='ocr:document_version_ocr_download'
)
link_document_version_ocr_errors_list = Link(
    args='resolved_object.id',
    icon=icon_document_version_ocr_errors_list,
    permissions=(permission_document_version_ocr_content_view,),
    text=_('OCR errors'), view='ocr:document_version_ocr_error_list'
)
link_document_version_ocr_submit_multiple = Link(
    icon=icon_document_version_ocr_submit_multiple,
    text=_('Submit for OCR'), view='ocr:document_version_ocr_submit_multiple'
)
link_document_version_ocr_submit_single = Link(
    args='resolved_object.id',
    icon=icon_document_version_ocr_submit_single,
    permissions=(permission_document_version_ocr,), text=_('Submit for OCR'),
    view='ocr:document_version_ocr_submit_single'
)

# Document version page

link_document_version_page_ocr_content_detail_view = Link(
    args='resolved_object.id',
    icon=icon_document_version_page_ocr_content_detail_view,
    permissions=(permission_document_version_ocr_content_view,),
    text=_('OCR'), view='ocr:document_version_page_ocr_content_detail_view'
)
link_document_version_page_ocr_content_edit_view = Link(
    args='resolved_object.id',
    icon=icon_document_version_page_ocr_content_edit_view,
    permissions=(permission_document_version_ocr_content_edit,),
    text=_('Edit OCR'), view='ocr:document_version_page_ocr_content_edit_view'
)

# Other

link_entry_list = Link(
    icon=icon_entry_list,
    permissions=(permission_document_version_ocr,), text=_('OCR errors'),
    view='ocr:entry_list'
)



from django.db import models, migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '__first__'),
    ]

    operations = [
        migrations.CreateModel(
            name='DocumentMetadata',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False,
                        auto_created=True, primary_key=True
                    )
                ),
                (
                    'value', models.CharField(
                        db_index=True, max_length=255, null=True,
                        verbose_name='Value', blank=True
                    )
                ),
                (
                    'document', models.ForeignKey(
                        on_delete=models.CASCADE, related_name='metadata',
                        to='documents.Document', verbose_name='Document'
                    )
                ),
            ],
            options={
                'verbose_name': 'Document metadata',
                'verbose_name_plural': 'Document metadata',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='DocumentTypeMetadataType',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False,
                        auto_created=True, primary_key=True
                    )
                ),
                (
                    'required', models.BooleanField(
                        default=False, verbose_name='Required'
                    )
                ),
                (
                    'document_type', models.ForeignKey(
                        on_delete=models.CASCADE, related_name='metadata',
                        to='documents.DocumentType',
                        verbose_name='Document type'
                    )
                ),
            ],
            options={
                'verbose_name': 'Document type metadata type options',
                'verbose_name_plural': 'Document type metadata types options',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='MetadataType',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False,
                        auto_created=True, primary_key=True
                    )
                ),
                (
                    'name', models.CharField(
                        help_text='Do not use python reserved words, '
                        'or spaces.', unique=True, max_length=48,
                        verbose_name='Name'
                    )
                ),
                (
                    'title', models.CharField(
                        max_length=48, verbose_name='Title'
                    )
                ),
                (
                    'default', models.CharField(
                        help_text='Enter a string to be evaluated.',
                        max_length=128, null=True, verbose_name='Default',
                        blank=True
                    )
                ),
                (
                    'lookup', models.TextField(
                        help_text='Enter a string to be evaluated that '
                        'returns an iterable.', null=True,
                        verbose_name='Lookup', blank=True)
                ),
                (
                    'validation', models.CharField(
                        blank=True, max_length=64,
                        verbose_name='Validation function name',
                        choices=[
                            ('Parse date', 'Parse date'),
                            ('Parse date and time', 'Parse date and time'),
                            ('Parse time', 'Parse time')
                        ]
                    )
                ),
            ],
            options={
                'ordering': ('title',),
                'verbose_name': 'Metadata type',
                'verbose_name_plural': 'Metadata types',
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='documenttypemetadatatype',
            name='metadata_type',
            field=models.ForeignKey(
                on_delete=models.CASCADE, to='metadata.MetadataType',
                verbose_name='Metadata type'
            ),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='documenttypemetadatatype',
            unique_together=set([('document_type', 'metadata_type')]),
        ),
        migrations.AddField(
            model_name='documentmetadata',
            name='metadata_type',
            field=models.ForeignKey(
                on_delete=models.CASCADE, to='metadata.MetadataType',
                verbose_name='Type'
            ),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='documentmetadata',
            unique_together=set([('document', 'metadata_type')]),
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(label=_('Search'), name='search')

permission_search_tools = namespace.add_permission(
    label=_('Execute search tools'), name='search_tools'
)



from django.db.models import Q

from mayan.apps.events.classes import EventType

from ...models import (
    WorkflowTransition, WorkflowTransitionField,
    WorkflowTransitionTriggerEvent
)

from ..literals import (
    TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_HELP_TEXT,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL_EDITED,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_NAME,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_TYPE,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED
)


class WorkflowTransitionFieldViewTestMixin:
    def _request_workflow_template_transition_field_create_view(self):
        pk_list = list(
            WorkflowTransitionField.objects.values_list('pk', flat=True)
        )

        response = self.post(
            viewname='document_states:workflow_template_transition_field_create',
            kwargs={
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }, data={
                'field_type': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_TYPE,
                'name': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_NAME,
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL,
                'help_text': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_HELP_TEXT
            }
        )

        try:
            self._test_workflow_template_transition_field = WorkflowTransitionField.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowTransitionField.DoesNotExist:
            self._test_workflow_template_transition_field = None

        return response

    def _request_workflow_template_transition_field_delete_view(self):
        return self.post(
            viewname='document_states:workflow_template_transition_field_delete',
            kwargs={
                'workflow_template_transition_field_id': self._test_workflow_template_transition_field.pk
            }
        )

    def _request_workflow_template_transition_field_edit_view(self):
        return self.post(
            viewname='document_states:workflow_template_transition_field_edit',
            kwargs={
                'workflow_template_transition_field_id': self._test_workflow_template_transition_field.pk
            }, data={
                'field_type': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_TYPE,
                'name': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_NAME,
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL_EDITED,
                'help_text': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_HELP_TEXT
            }
        )

    def _request_test_workflow_template_transition_field_list_view(self):
        return self.get(
            viewname='document_states:workflow_template_transition_field_list',
            kwargs={
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }
        )


class WorkflowTemplateTransitionAPIViewTestMixin:
    def _request_test_workflow_template_transition_create_api_view(
        self, extra_data=None
    ):
        data = {
            'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL,
            'origin_state_id': self._test_workflow_template_states[0].pk,
            'destination_state_id': self._test_workflow_template_states[1].pk
        }

        if extra_data:
            data.update(extra_data)

        pk_list = list(
            WorkflowTransition.objects.values_list('pk', flat=True)
        )

        response = self.post(
            viewname='rest_api:workflow-template-transition-list', kwargs={
                'workflow_template_id': self._test_workflow_template.pk
            }, data=data
        )

        try:
            self._test_workflow_template_transition = WorkflowTransition.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowTransition.DoesNotExist:
            self._test_workflow_template_transition = None

        return response

    def _request_test_workflow_template_transition_delete_api_view(self):
        return self.delete(
            viewname='rest_api:workflow-template-transition-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }
        )

    def _request_test_workflow_template_transition_detail_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }
        )

    def _request_test_workflow_template_transition_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-list',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_transition_edit_patch_api_view(self):
        return self.patch(
            viewname='rest_api:workflow-template-transition-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED,
                'origin_state_id': self._test_workflow_template_states[1].pk,
                'destination_state_id': self._test_workflow_template_states[0].pk
            }
        )

    def _request_test_workflow_template_transition_edit_put_api_view_via(self):
        return self.put(
            viewname='rest_api:workflow-template-transition-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED,
                'origin_state_id': self._test_workflow_template_states[1].pk,
                'destination_state_id': self._test_workflow_template_states[0].pk
            }
        )


class WorkflowTransitionFieldAPIViewTestMixin:
    def _request_test_workflow_template_transition_field_create_api_view(self):
        pk_list = list(WorkflowTransitionField.objects.values_list('pk'))

        response = self.post(
            viewname='rest_api:workflow-template-transition-field-list',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk,
            }, data={
                'field_type': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_TYPE,
                'name': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_NAME,
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL,
                'help_text': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_HELP_TEXT
            }
        )

        try:
            self._test_workflow_template_transition_field = WorkflowTransitionField.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowTransitionField.DoesNotExist:
            self._test_workflow_template_transition_field = None

        return response

    def _request_test_workflow_template_transition_field_delete_api_view(self):
        return self.delete(
            viewname='rest_api:workflow-template-transition-field-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk,
                'workflow_template_transition_field_id': self._test_workflow_template_transition_field.pk
            }
        )

    def _request_test_workflow_template_transition_field_detail_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-field-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk,
                'workflow_template_transition_field_id': self._test_workflow_template_transition_field.pk
            }
        )

    def _request_test_workflow_template_transition_field_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:workflow-template-transition-field-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk,
                'workflow_template_transition_field_id': self._test_workflow_template_transition_field.pk
            }, data={
                'label': '{} edited'.format(
                    self._test_workflow_template_transition_field
                )
            }
        )

    def _request_test_workflow_template_transition_field_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-field-list',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }
        )


class WorkflowTransitionFieldTestMixin:
    def _create_test_workflow_template_transition_field(self, extra_data=None):
        kwargs = {
            'field_type': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_TYPE,
            'name': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_NAME,
            'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_LABEL,
            'help_text': TEST_WORKFLOW_TEMPLATE_TRANSITION_FIELD_HELP_TEXT
        }
        kwargs.update(extra_data or {})

        self._test_workflow_template_transition_field = self._test_workflow_template_transition.fields.create(
            **kwargs
        )


class WorkflowTemplateTransitionTriggerAPIViewTestMixin:
    def _request_test_workflow_template_transition_trigger_create_api_view(self):
        data = {
            'event_type_id': self._test_event_type.id
        }

        pk_list = list(
            WorkflowTransitionTriggerEvent.objects.values_list(
                'pk', flat=True
            )
        )

        response = self.post(
            viewname='rest_api:workflow-template-transition-trigger-list',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }, data=data
        )

        try:
            self._test_workflow_template_transition_trigger = WorkflowTransitionTriggerEvent.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowTransitionTriggerEvent.DoesNotExist:
            self._test_workflow_template_transition_trigger = None

        return response

    def _request_test_workflow_template_transition_trigger_delete_api_view(self):
        return self.delete(
            viewname='rest_api:workflow-template-transition-trigger-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk,
                'workflow_template_transition_trigger_id': self._test_workflow_template_transition_trigger.pk
            }
        )

    def _request_test_workflow_template_transition_trigger_detail_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-trigger-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk,
                'workflow_template_transition_trigger_id': self._test_workflow_template_transition_trigger.pk
            }
        )

    def _request_test_workflow_template_transition_trigger_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-transition-trigger-list',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }
        )

    def _request_test_workflow_template_transition_trigger_edit_patch_api_view(self):
        return self.patch(
            viewname='rest_api:workflow-template-transition-trigger-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk,
                'workflow_template_transition_trigger_id': self._test_workflow_template_transition_trigger.pk
            }, data={
                'event_type_id': self._test_event_type.id
            }
        )

    def _request_test_workflow_template_transition_trigger_edit_put_api_view(self):
        return self.put(
            viewname='rest_api:workflow-template-transition-trigger-detail',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk,
                'workflow_template_transition_id': self._test_workflow_template_transition.pk,
                'workflow_template_transition_trigger_id': self._test_workflow_template_transition_trigger.pk
            }, data={
                'event_type_id': self._test_event_type.id
            }
        )


class WorkflowTemplateTransitionTriggerTestMixin:
    def setUp(self):
        super().setUp()
        self._test_workflow_template_transition_triggers = []

    def _create_test_workflow_template_transition_trigger(self):
        event_type = EventType.get(id=self._test_event_type.id)

        self._test_workflow_template_transition_trigger = self._test_workflow_template_transition.trigger_events.create(
            event_type=event_type.get_stored_event_type()
        )
        self._test_workflow_template_transition_triggers.append(
            self._test_workflow_template_transition_trigger
        )


class WorkflowTemplateTransitionTriggerViewTestMixin:
    def _request_test_workflow_template_transition_event_list_view(self):
        return self.get(
            viewname='document_states:workflow_template_transition_triggers',
            kwargs={
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }
        )


class WorkflowTemplateTransitionViewTestMixin:
    def _request_test_workflow_template_transition_create_view(self):
        pk_list = list(
            WorkflowTransition.objects.values_list('pk', flat=True)
        )

        response = self.post(
            viewname='document_states:workflow_template_transition_create',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL,
                'origin_state': self._test_workflow_template_states[0].pk,
                'destination_state': self._test_workflow_template_states[1].pk
            }
        )

        try:
            self._test_workflow_template_transition = WorkflowTransition.objects.get(
                ~Q(pk__in=pk_list)
            )
        except WorkflowTransition.DoesNotExist:
            self._test_workflow_template_transition = None

        return response

    def _request_test_workflow_template_transition_delete_view(self):
        return self.post(
            viewname='document_states:workflow_template_transition_delete',
            kwargs={
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }
        )

    def _request_test_workflow_template_transition_edit_view(self):
        return self.post(
            viewname='document_states:workflow_template_transition_edit',
            kwargs={
                'workflow_template_transition_id': self._test_workflow_template_transition.pk
            }, data={
                'label': TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED,
                'origin_state': self._test_workflow_template_states[0].pk,
                'destination_state': self._test_workflow_template_states[1].pk
            }
        )

    def _request_test_workflow_template_transition_list_view(self):
        return self.get(
            viewname='document_states:workflow_template_transition_list',
            kwargs={
                'workflow_template_id': self._test_workflow_template.pk
            }
        )



from collections import deque
import logging

import elasticsearch
from elasticsearch import Elasticsearch, helpers
from elasticsearch_dsl import Q, Search

from mayan.apps.common.utils import parse_range

from ..classes import SearchBackend, SearchModel
from ..exceptions import DynamicSearchException
from ..settings import setting_results_limit

from .literals import (
    DEFAULT_ELASTICSEARCH_CLIENT_MAXSIZE,
    DEFAULT_ELASTICSEARCH_CLIENT_SNIFF_ON_START,
    DEFAULT_ELASTICSEARCH_CLIENT_SNIFF_ON_CONNECTION_FAIL,
    DEFAULT_ELASTICSEARCH_CLIENT_SNIFFER_TIMEOUT, DEFAULT_ELASTICSEARCH_HOST,
    DEFAULT_ELASTICSEARCH_INDICES_NAMESPACE,
    DJANGO_TO_ELASTICSEARCH_FIELD_MAP
)

logger = logging.getLogger(name=__name__)


class ElasticSearchBackend(SearchBackend):
    _client = None
    _search_model_mappings = {}
    field_map = DJANGO_TO_ELASTICSEARCH_FIELD_MAP

    def __init__(self, **kwargs):
        self.client_kwargs = {}

        self.indices_namespace = kwargs.pop(
            'indices_namespace', DEFAULT_ELASTICSEARCH_INDICES_NAMESPACE
        )

        host = kwargs.pop('client_host', DEFAULT_ELASTICSEARCH_HOST)
        hosts = kwargs.pop('client_hosts', None)

        if not hosts:
            hosts = (host,)

        self.client_kwargs['hosts'] = hosts

        self.client_kwargs['http_auth'] = kwargs.pop('client_http_auth', None)
        self.client_kwargs['port'] = kwargs.pop('client_port', None)
        self.client_kwargs['scheme'] = kwargs.pop('client_scheme', None)

        self.client_kwargs['maxsize'] = kwargs.pop(
            'client_maxsize', DEFAULT_ELASTICSEARCH_CLIENT_MAXSIZE
        )
        self.client_kwargs['sniff_on_start'] = kwargs.pop(
            'client_sniff_on_start', DEFAULT_ELASTICSEARCH_CLIENT_SNIFF_ON_START
        )
        self.client_kwargs['sniff_on_connection_fail'] = kwargs.pop(
            'client_sniff_on_connection_fail', DEFAULT_ELASTICSEARCH_CLIENT_SNIFF_ON_CONNECTION_FAIL
        )
        self.client_kwargs['sniffer_timeout'] = kwargs.pop(
            'client_sniffer_timeout', DEFAULT_ELASTICSEARCH_CLIENT_SNIFFER_TIMEOUT
        )

        super().__init__(**kwargs)

    def _get_status(self):
        client = self.get_client()
        result = []

        title = 'Elastic Search search model indexing status'
        result.append(title)
        result.append(len(title) * '=')

        stats = client.indices.stats()

        for search_model in SearchModel.all():
            index_name = self.get_index_name(search_model=search_model)
            index_stats = stats['indices'][index_name]

            result.append(
                '{}: {}'.format(
                    search_model.label, index_stats['total']['docs']['count']
                )
            )

        return '\n'.join(result)

    def _initialize(self):
        self.update_mappings()

    def _search(
        self, query, search_model, user, global_and_search=False,
        ignore_limit=False
    ):
        client = self.get_client()
        index_name = self.get_index_name(
            search_model=search_model
        )

        search = Search(index=index_name, using=client)

        final_elasticsearch_query = None

        for key, value in query.items():
            elasticsearch_query = Q(
                Q(
                    name_or_query='fuzzy', _expand__to_dot=False, **{key: value}
                ) | Q(
                    name_or_query='match', _expand__to_dot=False, **{key: value}
                ) | Q(
                    name_or_query='regexp', _expand__to_dot=False, **{key: value}
                ) | Q(
                    name_or_query='wildcard', _expand__to_dot=False, **{key: value}
                )
            )

            if final_elasticsearch_query is None:
                final_elasticsearch_query = elasticsearch_query
            else:
                if global_and_search:
                    final_elasticsearch_query &= elasticsearch_query
                else:
                    final_elasticsearch_query |= elasticsearch_query

        search = search.source(None).query(final_elasticsearch_query)

        client.indices.refresh(index=index_name)

        if ignore_limit:
            limit = None
        else:
            limit = setting_results_limit.value

        response = search[0:limit].execute()

        id_list = []

        for hit in response:
            id_list.append(hit['id'])

        return search_model.get_queryset().filter(
            pk__in=id_list
        ).distinct()

    def close(self):
        self.get_client().transport.close()
        self.__class__._client = None

    def deindex_instance(self, instance):
        search_model = SearchModel.get_for_model(instance=instance)
        client = self.get_client()
        client.delete(
            id=instance.pk,
            index=self.get_index_name(search_model=search_model)
        )

    def get_client(self):
        try:
            if self.__class__._client is None:
                self.__class__._client = Elasticsearch(**self.client_kwargs)
        except Exception as exception:
            raise DynamicSearchException(
                'Unable to instantiate client; {}'.format(self.client_kwargs)
            ) from exception

        return self.__class__._client

    def get_index_name(self, search_model):
        return '{}-{}'.format(
            self.indices_namespace, search_model.model_name.lower()
        )

    def get_search_model_mappings(self, search_model):
        try:
            return self.__class__._search_model_mappings[search_model]
        except KeyError:
            mappings = {}

            field_map = self.get_resolved_field_map(search_model=search_model)
            for field_name, search_field_data in field_map.items():
                mappings[field_name] = {'type': search_field_data['field'].name}

            self.__class__._search_model_mappings[search_model] = mappings
        return mappings

    def index_instance(self, instance, exclude_model=None, exclude_kwargs=None):
        search_model = SearchModel.get_for_model(instance=instance)

        document = search_model.populate(
            backend=self, instance=instance, exclude_model=exclude_model,
            exclude_kwargs=exclude_kwargs
        )

        self.get_client().index(
            index=self.get_index_name(search_model=search_model),
            id=instance.pk, document=document
        )

    def index_search_model(self, search_model, range_string=None):
        client = self.get_client()
        index_name = self.get_index_name(search_model=search_model)

        def generate_actions():
            queryset = search_model.get_queryset()

            if range_string:
                queryset = queryset.filter(pk__in=list(parse_range(range_string=range_string)))

            for instance in queryset:
                kwargs = search_model.populate(
                    backend=self, instance=instance
                )
                kwargs['_id'] = kwargs['id']

                yield kwargs

        bulk_indexing_generator = helpers.streaming_bulk(
            client=client, index=index_name, actions=generate_actions(),
            yield_ok=False
        )

        deque(iterable=bulk_indexing_generator, maxlen=0)

    def reset(self, search_model=None):
        self.tear_down(search_model=search_model)
        self.update_mappings(search_model=search_model)

    def tear_down(self, search_model=None):
        client = self.get_client()
        if search_model:
            client.indices.delete(
                index=self.get_index_name(search_model=search_model)
            )
        else:
            client.indices.delete(
                index='{}-*'.format(self.indices_namespace)
            )

    def update_mappings(self, search_model=None):
        client = self.get_client()

        if search_model:
            search_models = (search_model,)
        else:
            search_models = SearchModel.all()

        for search_model in search_models:
            index_name = self.get_index_name(search_model=search_model)

            mappings = self.get_search_model_mappings(search_model=search_model)

            try:
                client.indices.create(
                    index=index_name,
                    body={'mappings': {'properties': mappings}}
                )
            except elasticsearch.exceptions.RequestError:
                try:
                    client.indices.put_mapping(
                        index=index_name,
                        body={'properties': mappings}
                    )
                except elasticsearch.exceptions.RequestError:
                    """There a mapping changes that were not allowed.
                    Example: Text to Keyword.
                    Boot up regardless and allow user to reindex to delete
                    old indices.
                    """



from django.utils import translation


def to_language(promise, language):
    current_language = translation.get_language()
    translation.activate(language=language)
    result = str(promise)
    translation.activate(language=current_language)

    return result



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('tags', '0006_documenttag'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='tag',
            options={
                'ordering': ('label',), 'verbose_name': 'Tag',
                'verbose_name_plural': 'Tags'
            },
        ),
    ]



from mayan.apps.testing.tests.base import GenericViewTestCase

from ..permissions import permission_task_view

from .mixins import TaskManagerTestMixin, TaskManagerViewTestMixin


class TaskManagerViewTestCase(
    TaskManagerTestMixin, TaskManagerViewTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_queue()

    def test_queue_list_view_no_permission(self):
        response = self._request_queue_list()

        self.assertEqual(response.status_code, 403)

    def test_queue_list_view_with_permissions(self):
        self.grant_permission(permission=permission_task_view)

        response = self._request_queue_list()
        self.assertContains(
            response, text=self.test_queue.name, status_code=200
        )



from mayan.apps.testing.tests.base import BaseTestCase

from ..events import event_cache_created, event_cache_purged
from ..models import Cache

from .mixins import CacheTestMixin


class CacheEventsTestCase(CacheTestMixin, BaseTestCase):
    def test_cache_create_event(self):
        self._clear_events()

        self._create_test_cache()

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        cache = Cache.objects.last()

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, cache)
        self.assertEqual(events[0].target, cache)
        self.assertEqual(events[0].verb, event_cache_created.id)

    def test_cache_purge_event(self):
        self._create_test_cache()

        self._clear_events()

        self.test_cache.purge()

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        cache = Cache.objects.last()

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, cache)
        self.assertEqual(events[0].target, cache)
        self.assertEqual(events[0].verb, event_cache_purged.id)



from django.utils.encoding import force_text

from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import (
    event_download_file_deleted, event_download_file_downloaded
)
from ..models import DownloadFile

from .literals import TEST_CONTENT
from .mixins import DownloadFileTestMixin, DownloadFileViewTestMixin


class DownloadFileViewTestCase(
    DownloadFileTestMixin, DownloadFileViewTestMixin, GenericViewTestCase
):
    def test_download_file_no_permission_delete_view(self):
        self._create_test_download_file()

        download_file_count = DownloadFile.objects.count()

        self._clear_events()

        response = self._request_test_download_file_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, None)
        self.assertEqual(events[0].verb, event_download_file_deleted.id)

    def test_download_file_no_permission_with_content_object_delete_view(self):
        self._create_test_object()
        DownloadFile.objects.register_content_object(model=self.TestModel)
        self._create_test_download_file(content_object=self.test_object)

        download_file_count = DownloadFile.objects.count()

        self._clear_events()

        response = self._request_test_download_file_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_object)
        self.assertEqual(events[0].verb, event_download_file_deleted.id)

    def test_download_file_with_permission_delete_view_no_permission(self):
        self._create_test_download_file_with_permission()

        download_file_count = DownloadFile.objects.count()

        self._clear_events()

        response = self._request_test_download_file_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_download_file_with_permission_delete_view_with_access(self):
        self._create_test_download_file_with_permission()

        self.grant_access(
            obj=self.test_download_file,
            permission=self.test_permission
        )

        download_file_count = DownloadFile.objects.count()

        self._clear_events()

        response = self._request_test_download_file_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, None)
        self.assertEqual(events[0].verb, event_download_file_deleted.id)

    def test_download_file_no_permission_download_view(self):
        # Set the expected_content_types for
        # common.tests.mixins.ContentTypeCheckMixin
        self.expected_content_types = ('text/plain',)

        self._create_test_download_file(content=TEST_CONTENT)

        self._clear_events()

        response = self._request_test_download_file_download_view()
        self.assertEqual(response.status_code, 200)

        with self.test_download_file.open(mode='r') as file_object:
            self.assert_download_response(
                response=response, content=file_object.read(),
                filename=force_text(s=self.test_download_file),
                mime_type='text/plain'
            )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_download_file)
        self.assertEqual(events[0].verb, event_download_file_downloaded.id)

    def test_download_file_with_permission_download_view_no_permission(self):
        self._create_test_download_file_with_permission(content=TEST_CONTENT)

        self._clear_events()

        response = self._request_test_download_file_download_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_download_file_with_permission_download_view_with_access(self):
        # Set the expected_content_types for
        # common.tests.mixins.ContentTypeCheckMixin
        self.expected_content_types = ('text/plain',)

        self._create_test_download_file_with_permission(content=TEST_CONTENT)

        self.grant_access(
            obj=self.test_download_file,
            permission=self.test_permission
        )

        self._clear_events()

        response = self._request_test_download_file_download_view()
        self.assertEqual(response.status_code, 200)

        with self.test_download_file.open(mode='r') as file_object:
            self.assert_download_response(
                response=response, content=file_object.read(),
                filename=force_text(s=self.test_download_file),
                mime_type='text/plain'
            )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_download_file)
        self.assertEqual(events[0].verb, event_download_file_downloaded.id)

    def test_download_file_no_permission_list_view(self):
        self._create_test_download_file()

        self._clear_events()

        response = self._request_test_download_file_list_view()

        self.assertContains(
            response=response, text=str(self.test_download_file),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_download_file_with_permission_list_view_no_permission(self):
        self._create_test_download_file_with_permission()

        self._clear_events()

        response = self._request_test_download_file_list_view()

        self.assertNotContains(
            response=response, text=str(self.test_download_file),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_download_file_with_permission_list_view_with_access(self):
        self._create_test_download_file_with_permission()

        self.grant_access(
            obj=self.test_download_file,
            permission=self.test_permission
        )

        self._clear_events()

        response = self._request_test_download_file_list_view()

        self.assertContains(
            response=response, text=str(self.test_download_file),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.contrib.contenttypes.models import ContentType

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.messaging.events import event_message_created
from mayan.apps.messaging.models import Message
from mayan.apps.storage.events import event_download_file_created
from mayan.apps.storage.models import DownloadFile

from ..events import event_events_exported
from ..permissions import permission_events_export

from .mixins import (
    EventsExportViewTestMixin, EventTestMixin, EventTypeTestMixin
)


class EventExportViewTestCase(
    EventTestMixin, EventTypeTestMixin, EventsExportViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_event_type()
        self._create_test_user()
        self.test_object = self.test_document_type

        content_type = ContentType.objects.get_for_model(
            model=self.test_object
        )

        self.view_arguments = {
            'app_label': content_type.app_label,
            'model_name': content_type.model,
            'object_id': self.test_object.pk
        }

        ModelPermission.register(
            model=self.test_object._meta.model, permissions=(
                permission_events_export,
            )
        )

    def test_events_list_export_view_no_permission(self):
        self._clear_events()

        self._create_test_event(target=self.test_object)

        response = self._request_test_events_list_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        self.assertEqual(events.first(), self.test_event)

        self.assertEqual(
            events[0].action_object, self.test_event.action_object
        )
        self.assertEqual(events[0].actor, self.test_event.actor)
        self.assertEqual(events[0].target, self.test_event.target)
        self.assertEqual(events[0].verb, self.test_event.verb)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_download_file)
        self.assertEqual(events[1].verb, event_download_file_created.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_events_exported.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, test_message)
        self.assertEqual(events[3].target, test_message)
        self.assertEqual(events[3].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() not in file_object.read()
            )

    def test_events_list_export_view_with_access(self):
        self.grant_access(
            obj=self.test_object, permission=permission_events_export
        )

        self._clear_events()

        self._create_test_event(target=self.test_object)

        response = self._request_test_events_list_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        self.assertEqual(
            events[0].action_object, self.test_event.action_object
        )
        self.assertEqual(events[0].actor, self.test_event.actor)
        self.assertEqual(events[0].target, self.test_event.target)
        self.assertEqual(events[0].verb, self.test_event.verb)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_download_file)
        self.assertEqual(events[1].verb, event_download_file_created.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_events_exported.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, test_message)
        self.assertEqual(events[3].target, test_message)
        self.assertEqual(events[3].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() in file_object.read()
            )

    def test_events_for_object_export_view_no_permission(self):
        self._clear_events()

        self._create_test_event(
            actor=self.test_user, action_object=self.test_object
        )

        response = self._request_events_for_object_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        self.assertEqual(
            events[0].action_object, self.test_event.action_object
        )
        self.assertEqual(events[0].actor, self.test_event.actor)
        self.assertEqual(events[0].target, self.test_event.target)
        self.assertEqual(events[0].verb, self.test_event.verb)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_download_file)
        self.assertEqual(events[1].verb, event_download_file_created.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_events_exported.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, test_message)
        self.assertEqual(events[3].target, test_message)
        self.assertEqual(events[3].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() not in file_object.read()
            )

    def test_events_for_object_export_view_with_access(self):
        self.grant_access(
            obj=self.test_object, permission=permission_events_export
        )

        self._clear_events()

        self._create_test_event(
            actor=self.test_user, action_object=self.test_object
        )

        response = self._request_events_for_object_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        self.assertEqual(
            events[0].action_object, self.test_event.action_object
        )
        self.assertEqual(events[0].actor, self.test_event.actor)
        self.assertEqual(events[0].target, self.test_event.target)
        self.assertEqual(events[0].verb, self.test_event.verb)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_download_file)
        self.assertEqual(events[1].verb, event_download_file_created.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_events_exported.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, test_message)
        self.assertEqual(events[3].target, test_message)
        self.assertEqual(events[3].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() in file_object.read()
            )

    def test_events_by_verb_export_view_no_permission(self):
        self._clear_events()

        self._create_test_event(
            actor=self.test_user, action_object=self.test_object
        )

        response = self._request_test_events_by_verb_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        self.assertEqual(
            events[0].action_object, self.test_event.action_object
        )
        self.assertEqual(events[0].actor, self.test_event.actor)
        self.assertEqual(events[0].target, self.test_event.target)
        self.assertEqual(events[0].verb, self.test_event.verb)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_download_file)
        self.assertEqual(events[1].verb, event_download_file_created.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_events_exported.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, test_message)
        self.assertEqual(events[3].target, test_message)
        self.assertEqual(events[3].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() not in file_object.read()
            )

    def test_events_by_verb_view_export_with_access(self):
        self.grant_access(
            obj=self.test_object, permission=permission_events_export
        )

        self._clear_events()

        self._create_test_event(
            actor=self.test_user, action_object=self.test_object
        )

        response = self._request_test_events_by_verb_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        self.assertEqual(
            events[0].action_object, self.test_event.action_object
        )
        self.assertEqual(events[0].actor, self.test_event.actor)
        self.assertEqual(events[0].target, self.test_event.target)
        self.assertEqual(events[0].verb, self.test_event.verb)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_download_file)
        self.assertEqual(events[1].verb, event_download_file_created.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_events_exported.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, test_message)
        self.assertEqual(events[3].target, test_message)
        self.assertEqual(events[3].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() in file_object.read()
            )


class CurrentUsetEventExportViewTestCase(
    EventTestMixin, EventTypeTestMixin, EventsExportViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_event_type()
        self._create_test_user()
        self.test_object = self._test_case_user

        content_type = ContentType.objects.get_for_model(
            model=self.test_object
        )

        self.view_arguments = {
            'app_label': content_type.app_label,
            'model_name': content_type.model,
            'object_id': self.test_object.pk
        }

    def test_current_user_events_export_view_no_permission(self):
        self._clear_events()

        self._create_test_event(
            actor=self._test_case_user, action_object=self.test_object
        )

        response = self._request_events_for_object_export_view()
        self.assertNotContains(
            response=response, text=str(self._test_event_type), status_code=302
        )

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        self.assertEqual(
            events[0].action_object, self.test_event.action_object
        )
        self.assertEqual(events[0].actor, self.test_event.actor)
        self.assertEqual(events[0].target, self.test_event.target)
        self.assertEqual(events[0].verb, self.test_event.verb)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_download_file)
        self.assertEqual(events[1].verb, event_download_file_created.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_events_exported.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, test_message)
        self.assertEqual(events[3].target, test_message)
        self.assertEqual(events[3].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() not in file_object.read()
            )

    def test_current_user_events_export_view_with_access(self):
        self.grant_access(
            obj=self._test_case_user, permission=permission_events_export
        )

        self._clear_events()

        self._create_test_event(
            actor=self._test_case_user, action_object=self.test_object
        )

        response = self._request_events_for_object_export_view()
        self.assertEqual(response.status_code, 302)

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        self.assertEqual(
            events[0].action_object, self.test_event.action_object
        )
        self.assertEqual(events[0].actor, self.test_event.actor)
        self.assertEqual(events[0].target, self.test_event.target)
        self.assertEqual(events[0].verb, self.test_event.verb)

        self.assertEqual(events[1].action_object, None)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_download_file)
        self.assertEqual(events[1].verb, event_download_file_created.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_download_file)
        self.assertEqual(events[2].verb, event_events_exported.id)

        self.assertEqual(events[3].action_object, None)
        self.assertEqual(events[3].actor, test_message)
        self.assertEqual(events[3].target, test_message)
        self.assertEqual(events[3].verb, event_message_created.id)

        with test_download_file.open() as file_object:
            self.assertTrue(
                str(self.test_object).encode() in file_object.read()
            )



from django.test import override_settings

from ..events import event_document_type_changed, event_document_viewed
from ..permissions import (
    permission_document_properties_edit, permission_document_view
)

from .base import GenericDocumentViewTestCase
from .mixins.document_mixins import DocumentViewTestMixin


class DocumentViewTestCase(
    DocumentViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_document_properties_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_properties_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_properties_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_document_properties_view()
        self.assertContains(
            response=response, status_code=200, text=self.test_document.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_properties_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_properties_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_properties_edit_get_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_properties_edit_get_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_properties_edit_get_view_with_access(self):
        self.grant_access(
            permission=permission_document_properties_edit,
            obj=self.test_document_type
        )

        self._clear_events()

        response = self._request_test_document_properties_edit_get_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_properties_edit_get_view_with_access(self):
        self.grant_access(
            permission=permission_document_properties_edit,
            obj=self.test_document_type
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_properties_edit_get_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    @override_settings(DOCUMENTS_LANGUAGE='fra')
    def test_document_properties_view_setting_non_us_language_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_document_properties_view()
        self.assertContains(
            response=response, status_code=200, text=self.test_document.label
        )
        self.assertContains(
            response=response, status_code=200,
            text='Language:</label>\n                \n                \n                    English'
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    @override_settings(DOCUMENTS_LANGUAGE='fra')
    def test_document_properties_edit_get_view_setting_non_us_language_with_access(self):
        self.grant_access(
            permission=permission_document_properties_edit,
            obj=self.test_document_type
        )

        self._clear_events()

        response = self._request_test_document_properties_edit_get_view()
        self.assertContains(
            response=response, status_code=200,
            text='<option value="eng" selected>English</option>',
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_list_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_list_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(response.context['object_list'].count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_list_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_document_list_view()
        self.assertContains(
            response=response, status_code=200, text=self.test_document.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_list_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_list_view()
        self.assertNotContains(
            response=response, status_code=200, text=self.test_document.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_change_post_view_no_permission(self):
        self._create_test_document_type()

        document_type = self.test_document.document_type

        self._clear_events()

        response = self._request_test_document_type_change_post_view()
        self.assertEqual(response.status_code, 404)

        self.test_document.refresh_from_db()
        self.assertEqual(
            self.test_document.document_type, document_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_change_post_view_with_access(self):
        self._create_test_document_type()

        document_type = self.test_document.document_type

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_properties_edit
        )

        self._clear_events()

        response = self._request_test_document_type_change_post_view()
        self.assertEqual(response.status_code, 302)

        self.test_document.refresh_from_db()
        self.assertNotEqual(
            self.test_document.document_type, document_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_types[1])
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document)
        self.assertEqual(events[0].verb, event_document_type_changed.id)

    def test_trashed_document_document_type_change_post_view_with_access(self):
        self._create_test_document_type()

        document_type = self.test_document.document_type

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_properties_edit
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_type_change_post_view()
        self.assertEqual(response.status_code, 404)

        self.test_document.refresh_from_db()
        self.assertEqual(
            self.test_document.document_type, document_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_change_view_get_no_permission(self):
        self._create_test_document_type()

        document_type = self.test_document.document_type

        self._clear_events()

        response = self._request_test_document_type_change_get_view()
        self.assertEqual(response.status_code, 404)

        self.test_document.refresh_from_db()
        self.assertEqual(
            self.test_document.document_type, document_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_change_view_get_with_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_properties_edit
        )

        self._clear_events()

        response = self._request_test_document_type_change_get_view()
        self.assertEqual(response.status_code, 200)

        self.test_document.refresh_from_db()
        self.assertEqual(
            self.test_document.document_type, self.test_document_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_type_change_view_get_with_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_properties_edit
        )

        self._create_test_document_type()

        document_type = self.test_document.document_type

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_type_change_get_view()
        self.assertEqual(response.status_code, 404)

        self.test_document.refresh_from_db()
        self.assertEqual(
            self.test_document.document_type, document_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_document_type_change_view_no_permission(self):
        self._create_test_document_type()

        document_type = self.test_document.document_type

        self._clear_events()

        response = self._request_test_document_multiple_type_change()
        self.assertEqual(response.status_code, 404)

        self.test_document.refresh_from_db()
        self.assertEqual(
            self.test_document.document_type, document_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_document_type_change_view_with_permission(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_properties_edit
        )

        self._create_test_document_type()

        document_type = self.test_document.document_type

        self._clear_events()

        response = self._request_test_document_multiple_type_change()
        self.assertEqual(response.status_code, 302)

        self.test_document.refresh_from_db()
        self.assertNotEqual(
            self.test_document.document_type, document_type
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_types[1])
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document)
        self.assertEqual(events[0].verb, event_document_type_changed.id)

    def test_document_preview_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_preview_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_preview_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_document_preview_view()
        self.assertContains(
            response=response, status_code=200, text=self.test_document.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document)
        self.assertEqual(events[0].verb, event_document_viewed.id)

    def test_trashed_document_preview_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_preview_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import logging

from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.user_management.tests.mixins import GroupTestMixin

from ..exceptions import QuotaExceeded
from ..quota_backends import DocumentCountQuota, DocumentSizeQuota


class DocumentCountQuotaTestCase(GroupTestMixin, GenericDocumentTestCase):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        # Increase the initial usage count to 1 by uploading a document
        # as the test case user.
        self._upload_test_document(_user=self._test_case_user)
        self.test_case_silenced_logger_new_level = logging.FATAL + 10
        self._silence_logger(name='mayan.apps.documents.models')

    def test_user_all_document_type_all(self):
        self.test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=True,
            user_ids=(),
        )

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document()

    def test_user_all_document_type_all_two_users(self):
        self.test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=True,
            user_ids=(),
        )
        self._create_test_user()

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document(_user=self.test_user)

    def test_user_all_document_type_test(self):
        self.test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=False,
            document_type_ids=(self.test_document_type.pk,),
            group_ids=(),
            user_all=True,
            user_ids=(),
        )

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document()

    def test_user_test_document_type_all(self):
        self.test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=False,
            user_ids=(self._test_case_user.pk,),
        )

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document(_user=self._test_case_user)

    def test_group_test_document_type_all(self):
        self._create_test_group()
        self._test_case_user.groups.add(self.test_group)

        self.test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(self.test_group.pk,),
            user_all=False,
            user_ids=(),
        )

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document(_user=self._test_case_user)

    def test_allow(self):
        self.test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=False,
            user_ids=(),
        )

        self._upload_test_document(_user=self._test_case_user)

    def test_super_user_restriction(self):
        self._create_test_superuser()

        self.test_quota = DocumentCountQuota.create(
            documents_limit=1,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=True,
            user_ids=(),
        )

        self._upload_test_document(_user=self.test_superuser)


class DocumentSizeQuotaTestCase(GroupTestMixin, GenericDocumentTestCase):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self.test_case_silenced_logger_new_level = logging.FATAL + 10
        self._silence_logger(name='mayan.apps.documents.models')

    def test_user_all_document_type_all(self):
        self.test_quota = DocumentSizeQuota.create(
            document_size_limit=0.01,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=True,
            user_ids=(),
        )

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document()

    def test_user_all_document_type_test(self):
        self.test_quota = DocumentSizeQuota.create(
            document_size_limit=0.01,
            document_type_all=False,
            document_type_ids=(self.test_document_type.pk,),
            group_ids=(),
            user_all=True,
            user_ids=(),
        )

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document()

    def test_user_test_document_type_test(self):
        self.test_quota = DocumentSizeQuota.create(
            document_size_limit=0.01,
            document_type_all=False,
            document_type_ids=(self.test_document_type.pk,),
            group_ids=(),
            user_all=False,
            user_ids=(self._test_case_user.pk,),
        )

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document(_user=self._test_case_user)

    def test_group_test_document_type_test(self):
        self._create_test_group()
        self._test_case_user.groups.add(self.test_group)

        self.test_quota = DocumentSizeQuota.create(
            document_size_limit=0.01,
            document_type_all=False,
            document_type_ids=(self.test_document_type.pk,),
            group_ids=(self.test_group.pk,),
            user_all=False,
            user_ids=(),
        )

        with self.assertRaises(expected_exception=QuotaExceeded):
            self._upload_test_document(_user=self._test_case_user)

    def test_allow(self):
        self.test_quota = DocumentSizeQuota.create(
            document_size_limit=0.01,
            document_type_all=False,
            document_type_ids=(),
            group_ids=(),
            user_all=False,
            user_ids=(),
        )

        self._upload_test_document(_user=self._test_case_user)

    def test_super_user_restriction(self):
        self._create_test_superuser()

        self.test_quota = DocumentSizeQuota.create(
            document_size_limit=0.01,
            document_type_all=True,
            document_type_ids=(),
            group_ids=(),
            user_all=True,
            user_ids=(),
        )

        self._upload_test_document(_user=self.test_superuser)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(label=_('Statistics'), name='statistics')

permission_statistics_view = namespace.add_permission(
    label=_('View statistics'), name='statistics_view'
)



from django.utils.html import format_html_join

from mayan.apps.navigation.html_widgets import SourceColumnWidget


def widget_transition_events(transition):
    return format_html_join(
        sep='\n', format_string='<div class="">{}</div>', args_generator=(
            (
                transition_trigger.event_type.label,
            ) for transition_trigger in transition.trigger_events.all()
        )
    )


class WorkflowLogExtraDataWidget(SourceColumnWidget):
    template_name = 'document_states/extra_data.html'



import os
import tempfile

from django.conf import settings

DEFAULT_STORAGE_BACKEND = 'django.core.files.storage.FileSystemStorage'
DEFAULT_STORAGE_DOWNLOAD_FILE_STORAGE = 'django.core.files.storage.FileSystemStorage'
DEFAULT_STORAGE_DOWNLOAD_FILE_STORAGE_ARGUMENTS = {
    'location': os.path.join(settings.MEDIA_ROOT, 'download_files')
}
DEFAULT_STORAGE_SHARED_STORAGE = 'django.core.files.storage.FileSystemStorage'
DEFAULT_STORAGE_SHARED_STORAGE_ARGUMENTS = {
    'location': os.path.join(settings.MEDIA_ROOT, 'shared_files')
}
DEFAULT_STORAGE_TEMPORARY_DIRECTORY = tempfile.gettempdir()
DEFAULT_DOWNLOAD_FILE_EXPIRATION_INTERVAL = 60 * 24 * 2  # 2 days
DEFAULT_SHARED_UPLOADED_FILE_EXPIRATION_INTERVAL = 60 * 60 * 24 * 7  # 7 days

MSG_MIME_TYPES = (
    'application/vnd.ms-outlook', 'application/vnd.ms-office',
    'application/x-ole-storage'
)
STORAGE_NAME_DOWNLOAD_FILE = 'storage__downloadfile'
STORAGE_NAME_SHARED_UPLOADED_FILE = 'storage__shareduploadedfile'
TASK_DOWNLOAD_FILE_STALE_INTERVAL = 60 * 10  # 10 minutes
TASK_SHARED_UPLOADS_STALE_INTERVAL = 60 * 10  # 10 minutes



import logging

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.classes import (
    EventManagerMethodAfter, EventManagerSave
)
from mayan.apps.events.decorators import method_event
from mayan.apps.permissions.models import Role, StoredPermission

from .events import event_acl_created, event_acl_deleted, event_acl_edited
from .managers import AccessControlListManager

logger = logging.getLogger(name=__name__)


class AccessControlList(ExtraDataModelMixin, models.Model):
    """
    ACL means Access Control List it is a more fine-grained method of
    granting access to objects. In the case of ACLs, they grant access using
    3 elements: actor, permission, object. In this case the actor is the role,
    the permission is the Mayan permission and the object can be anything:
    a document, a folder, an index, etc. This means = "Grant X permissions
    to role Y for object Z". This model holds the permission, object, actor
    relationship for one access control list.
    Fields:
    * Role - Custom role that is being granted a permission. Roles are created
    in the Setup menu.
    """
    content_type = models.ForeignKey(
        on_delete=models.CASCADE, related_name='object_content_type',
        to=ContentType
    )
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey(
        ct_field='content_type', fk_field='object_id',
    )
    permissions = models.ManyToManyField(
        blank=True, related_name='acls', to=StoredPermission,
        verbose_name=_('Permissions')
    )
    role = models.ForeignKey(
        help_text=_(
            'Role to which the access is granted for the specified object.'
        ), on_delete=models.CASCADE, related_name='acls', to=Role,
        verbose_name=_('Role')
    )

    objects = AccessControlListManager()

    class Meta:
        ordering = ('pk',)
        unique_together = ('content_type', 'object_id', 'role')
        verbose_name = _('Access entry')
        verbose_name_plural = _('Access entries')

    def __str__(self):
        return _(
            'Role "%(role)s" permission\'s for "%(object)s"'
        ) % {
            'object': self.content_object,
            'role': self.role,
        }

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_acl_deleted,
        target='content_object'
    )
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    def get_absolute_url(self):
        return reverse(
            viewname='acls:acl_permissions', kwargs={'acl_id': self.pk}
        )

    def get_inherited_permissions(self):
        return AccessControlList.objects.get_inherited_permissions(
            obj=self.content_object, role=self.role
        )

    def permissions_add(self, queryset, _event_actor=None):
        for obj in queryset:
            self.permissions.add(obj)

        event_acl_edited.commit(
            action_object=self.content_object,
            actor=_event_actor or self._event_actor, target=self
        )

    def permissions_remove(self, queryset, _event_actor=None):
        for obj in queryset:
            self.permissions.remove(obj)

        event_acl_edited.commit(
            action_object=self.content_object,
            actor=_event_actor or self._event_actor, target=self
        )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'action_object': 'content_object',
            'event': event_acl_created,
            'target': 'self',
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class GlobalAccessControlListProxy(AccessControlList):
    class Meta:
        proxy = True
